import six

from pubtools.pulplib._impl import compat_attr as attr


# attr metadata private key for mapping between PulpObject attr and the corresponding
# field in Pulp (if it exists)
PULP2_FIELD = "_pubtools.pulplib.pulp2_field"

# attr metadata private key indicating whether a field is part of the unit_key.
PULP2_UNIT_KEY = "_pubtools.pulplib.pulp2_unit_key"

# attr metadata private key for converting from a pulp2 representation and a Python
# object.
# Why not using attr.ib built-in 'converter'?  Because that makes it public API,
# i.e. clients could construct PulpObject subclasses directly with Pulp data.
# This is not desired, better to have all Pulp conversion logic strictly hidden behind
# from_data methods.
PULP2_PY_CONVERTER = "_pubtools.pulplib.pulp2_to_py_converter"

# Inverse of the above: converter for Python value into Pulp2 value.
PY_PULP2_CONVERTER = "_pubtools.pulplib.py_to_pulp2_converter"


def pulp_attrib(
    pulp_field=None,
    pulp_py_converter=None,
    py_pulp_converter=None,
    unit_key=None,
    **kwargs
):
    """Drop-in replacement for attr.ib with added features:

    - applies a validator based on type automatically
    - supports pulplib-specific metadata via extra keyword arguments
    """
    metadata = kwargs.get("metadata") or {}

    if pulp_field:
        metadata[PULP2_FIELD] = pulp_field

    if pulp_py_converter:
        metadata[PULP2_PY_CONVERTER] = pulp_py_converter

    if py_pulp_converter:
        metadata[PY_PULP2_CONVERTER] = py_pulp_converter

    if unit_key is not None:
        metadata[PULP2_UNIT_KEY] = unit_key

    if "type" in kwargs:
        # As a convenience, you may define string types as type=str
        # on any python version, but what you'll actually get is
        # whatever's the primary string type (e.g. basestr on py2)
        if kwargs["type"] is str:
            kwargs["type"] = six.string_types[0]

        # If you haven't defined a validator, you get one automatically
        # for your requested type
        if "validator" not in kwargs:
            kwargs["validator"] = attr.validators.optional(
                attr.validators.instance_of(kwargs["type"])
            )

    kwargs["metadata"] = metadata
    return attr.ib(**kwargs)



from .common import PulpObject, DetachedException, InvalidDataException
from .repository import (
    Repository,
    YumRepository,
    FileRepository,
    ContainerImageRepository,
    PublishOptions,
    SyncOptions,
    FileSyncOptions,
    ContainerSyncOptions,
    YumSyncOptions,
)
from .unit import (
    Unit,
    FileUnit,
    ErratumUnit,
    ErratumReference,
    ErratumPackageCollection,
    ErratumPackage,
    ErratumModule,
    RpmUnit,
    RpmDependency,
    ModulemdUnit,
    ModulemdDependency,
    ModulemdDefaultsUnit,
    YumRepoMetadataFileUnit,
)
from .task import Task
from .distributor import Distributor
from .maintenance import MaintenanceReport, MaintenanceEntry



import os
import logging

from attr import validators
from frozenlist2 import frozenlist

from .base import Repository, SyncOptions, repo_type
from ...model.unit import FileUnit
from ..attr import pulp_attrib
from ... import compat_attr as attr


LOG = logging.getLogger("pubtools.pulplib")


@attr.s(kw_only=True, frozen=True)
class FileSyncOptions(SyncOptions):
    """Options controlling a file repository
    :meth:`~pubtools.pulplib.Repository.sync`.
    """

    remove_missing = pulp_attrib(default=False, type=bool)
    """If true, as the repository is synchronized, old files will be removed.
    """


@repo_type("iso-repo")
@attr.s(kw_only=True, frozen=True)
class FileRepository(Repository):
    """A :class:`~pubtools.pulplib.Repository` for generic file distribution."""

    # this class only overrides some defaults for attributes defined in super

    type = pulp_attrib(default="iso-repo", type=str, pulp_field="notes._repo-type")

    is_sigstore = attr.ib(
        default=attr.Factory(
            lambda self: self.id == "redhat-sigstore", takes_self=True
        ),
        type=bool,
        validator=validators.instance_of(bool),
    )

    mutable_urls = attr.ib(
        default=attr.Factory(lambda: frozenlist(["PULP_MANIFEST"])),
        type=list,
        converter=frozenlist,
    )

    def upload_file(self, file_obj, relative_url=None, **kwargs):
        """Upload a file to this repository.

        Args:
            file_obj (str, file object)
                If it's a string, then it's the path of a file to upload.

                Otherwise, it should be a
                `file-like object <https://docs.python.org/3/glossary.html#term-file-object>`_
                pointing at the bytes to upload. The client takes ownership
                of this file object; it should not be modified elsewhere,
                and will be closed when upload completes.

            relative_url (str)
                Path that should be used in remote repository, can either
                be a path to a directory or a path to a file, e.g:

                - if relative_url is 'foo/bar/' and file_obj has name 'f.txt',
                  the resulting remote path wll be 'foo/bar/f.txt'.

                - if relative_url is 'foo/bar/f.txt', no matter what the
                  name of file_obj is, the remote path is 'foo/bar/f.txt'.

                If omitted, the local name of the file will be used. Or,
                if file_obj is a file object without a `name` attribute,
                passing `relative_url` is mandatory.

            kwargs
                Additional field values to set on the uploaded unit.

                Any :class:`~pubtools.pulplib.FileUnit` fields documented as
                *mutable* may be included here (for example, ``cdn_path``).
                An error will occur if attempting to set other fields.

        Returns:
            Future[list of :class:`~pubtools.pulplib.Task`]
                A future which is resolved after content has been imported
                to this repo.

        Raises:
            DetachedException
                If this instance is not attached to a Pulp client.

        .. versionadded:: 1.2.0

        .. versionadded:: 2.20.0
            Added ability to set mutable fields on upload.
        """
        relative_url = self._get_relative_url(file_obj, relative_url)

        unit_key_fn = lambda upload: {
            "name": relative_url,
            "checksum": upload[0],
            "size": upload[1],
        }
        unit_metadata_fn = None

        usermeta = FileUnit._usermeta_from_kwargs(**kwargs)
        if usermeta:
            unit_metadata_fn = lambda _: usermeta

        return self._upload_then_import(
            file_obj, relative_url, "iso", unit_key_fn, unit_metadata_fn
        )

    def _get_relative_url(self, file_obj, relative_url):
        is_file_object = "close" in dir(file_obj)
        if not is_file_object:
            name = os.path.basename(file_obj)
            if not relative_url:
                relative_url = name
            elif relative_url.endswith("/"):
                relative_url = os.path.join(relative_url, name)
        elif is_file_object and (not relative_url or relative_url.endswith("/")):
            msg = "%s is missing a name attribute and relative_url was not provided"
            raise ValueError(msg % file_obj)

        return relative_url



from .base import Repository, PublishOptions, SyncOptions
from .container import ContainerImageRepository, ContainerSyncOptions
from .yum import YumRepository, YumSyncOptions
from .file import FileRepository, FileSyncOptions



import logging
import datetime

import jsonschema
import six

from more_executors.futures import f_map, f_proxy

from pubtools.pulplib._impl import compat_attr as attr
from pubtools.pulplib._impl.util import lookup, dict_put

from .attr import PULP2_FIELD
from .convert import get_converter

LOG = logging.getLogger("pubtools.pulplib")


class DetachedException(Exception):
    """If an operation is attempted on a Pulp object which requires an active client,
    and the object is not attached to any client, this exception is raised.
    """


class InvalidDataException(Exception):
    """Raised if raw Pulp data appears to be invalid (i.e. not matching expected schema)."""


class PulpObject(object):
    """Base class for all modeled Pulp objects.

    Instances of PulpObject subclasses may be obtained by get and search methods
    on :class:`~pubtools.pulplib.Client`, or may be instantiated directly by calls
    to :meth:`from_data` when the client is not used.

    Objects which are created via a client may be used to issue further requests
    to Pulp (for example, to update or delete the object).

    Pulp objects use `attrs <http://www.attrs.org/en/stable/>`_.
    Attributes are immutable. Helper functions such as :func:`attr.evolve`
    may be used to produce new instances.

    Attributes exposed on these Pulp objects include some generic attributes
    applicable to any Pulp installation, but also some custom attributes
    which only make sense for `release-engineering <https://github.com/release-engineering>`_
    Pulp servers.
    """

    # Classes should put a valid JSON schema here. This one will refuse to
    # validate anything.
    _SCHEMA = False

    @classmethod
    def from_data(cls, data):
        """Obtain a detached instance using data obtained from Pulp.

        This method is provided so that callers who are not using
        :class:`~pubtools.pulplib.Client` may make use of the Pulp object classes.

        This method must be invoked on the appropriate ``PulpObject`` subclass
        matching ``data``.  For example, ``Repository.from_data`` must be invoked
        with a repository object provided by Pulp's API.

        Args:
            data (dict)
                A dict containing a raw representation of a Pulp object, as rendered
                by Pulp's API.

        Returns:
            a new instance of ``cls``

        Raises:
            InvalidDataException
                If the provided ``data`` fails validation against an expected schema.

        Example:

            Opting-out of using the ``Client`` class and instead doing a plain ``requests.get``:

            .. code-block:: python

                url = 'https://pulp.example.com/pulp/api/v2/repositories/zoo/'
                data = requests.get(url).json()
                repo = Repository.from_data(data)
        """

        try:
            jsonschema.validate(instance=data, schema=cls._SCHEMA)

            kwargs = cls._data_to_init_args(data)
            return cls(**kwargs)

        except Exception as error:  # pylint:disable=broad-except
            LOG.exception(
                (
                    "An error occurred while loading Pulp data!\n"
                    "  Model class: %s\n"
                    "  Raw data:    %s"
                ),
                cls,
                repr(data),
            )

            msg = "%s.from_data invoked with invalid Pulp data", cls.__name__
            six.raise_from(InvalidDataException(msg), error)

    def _to_data(self):
        """Inverse of from_data: serialize a model object back to native Pulp form.

        This method is currently intended for internal use only.

        Returns:
            This object, in the native format used by pulp2 (i.e. some
            JSON-encodable type).
        """
        fields = attr.fields(type(self))

        out = {}
        for field in fields:
            pulp_field = field.metadata.get(PULP2_FIELD)

            if not pulp_field:
                # This field does not map to pulp
                continue

            python_value = getattr(self, field.name)

            # Note: in theory we should also get and use PY_PULP2_CONVERTER
            # here if that was set on the metadata. It's not currently
            # implemented because ErratumUnit is the only type for which
            # this code can be reached from public API, and it has no fields
            # which need a custom converter, so it would be dead code.
            # Implement it when you need it!
            #
            # For now, conversions supported by PulpObject are sufficient.
            pulp_value = PulpObject._any_to_data(python_value)

            # Put converted value into the output dict:
            # This may create nested dicts if needed, e.g. if
            # pulp_field is "notes.foobar", this will create a "notes"
            # dict in out if it does not already exist.
            dict_put(out, pulp_field, pulp_value)

        return out

    @classmethod
    def _any_to_data(cls, value):
        """Like the instance method _to_data, but also handles non-PulpObject values."""

        if isinstance(value, list):
            # Lists of objects are converted recursively.
            return [cls._any_to_data(elem) for elem in value]

        if isinstance(value, PulpObject):
            # It's a model object, then delegate to the instance method.
            return value._to_data()

        if isinstance(value, datetime.datetime):
            # For datetimes, we always use an ISO8601 timestamp format.
            return value.strftime("%Y-%m-%dT%H:%M:%SZ")

        # For anything else, we assume it can be used as-is.
        # strs and ints for example fall into this path.
        return value

    @classmethod
    def _data_to_init_args(cls, data):
        # maps from raw Pulp dict to a kwargs dict used to initialize
        # a new object of this class.
        #
        # The default implementation looks at defined attributes and metadata
        # (PULP2_FIELD).  If this is not sufficient, subclasses can override
        # this, and can also call super() to reuse this as needed.
        out = {}
        fields = attr.fields(cls)
        absent = object()

        for field in fields:
            pulp_field = field.metadata.get(PULP2_FIELD)
            if pulp_field:
                value = lookup(data, pulp_field, absent)
                if value is not absent:
                    converter = get_converter(field, value)
                    value = converter(value)
                    out[field.name] = value

        return out


@attr.s(kw_only=True, frozen=True)
class WithClient(object):
    # A mixin for objects holding a private reference to client.

    _client = attr.ib(default=None, init=False, repr=False, cmp=False, hash=False)

    def _set_client(self, client):
        self.__dict__["_client"] = client


class Deletable(WithClient):
    # A mixin for objects representing deletable resources.

    def __detach(self, retval):
        LOG.debug("Detaching %s after successful delete", self)
        self._set_client(None)
        return retval

    def _delete(self, resource_type, resource_id):
        client = self._client
        if not client:
            raise DetachedException()

        delete_f = client._delete_resource(resource_type, resource_id)
        delete_f = f_map(delete_f, self.__detach)
        return f_proxy(delete_f)



import random
import uuid
import threading
import hashlib
import json
import re

from collections import namedtuple

import six
from six.moves import StringIO

from more_executors.futures import f_return, f_return_error, f_flat_map, f_proxy

from pubtools.pulplib import (
    Page,
    PulpException,
    Criteria,
    Task,
    Repository,
    Distributor,
    Unit,
    FileUnit,
    MaintenanceReport,
)
from pubtools.pulplib._impl.client.search import search_for_criteria
from .. import compat_attr as attr

from .match import match_object
from . import units

Publish = namedtuple("Publish", ["repository", "tasks"])
Upload = namedtuple("Upload", ["repository", "tasks", "name", "sha256"])
Sync = namedtuple("Sync", ["repository", "tasks", "sync_config"])


class FakeClient(object):  # pylint:disable = too-many-instance-attributes
    # Client implementation holding data in memory rather than
    # using a remote Pulp server.
    #
    # This class is not public, but it must have all of the same public
    # API as the pubtools.pulplib.Client class. The idea is that any code
    # written against pubtools.pulplib.Client should be able to work with
    # an instance of this class swapped in.
    _PAGE_SIZE = 3
    _DEFAULT_TYPE_IDS = [
        "distribution",
        "drpm",
        "erratum",
        "iso",
        "modulemd_defaults",
        "modulemd",
        "package_category",
        "package_environment",
        "package_group",
        "package_langpacks",
        "repository",
        "rpm",
        "srpm",
        "yum_repo_metadata_file",
    ]

    def __init__(self):
        self._repositories = []

        # Similar to a real Pulp server, units are stored through a "unit key"
        # layer of indirection:
        # - repos contain unit keys
        # - a unit key refers to one (and only one) unit
        #
        # By storing data in this way, we ensure we have similar behavior and
        # constraints as a real Pulp server, e.g. cannot store two different units
        # which would have the same key; updating a unit referenced from multiple repos
        # effectively updates it in all repos at once; etc.
        self._repo_unit_keys = {}
        self._units_by_key = {}

        self._publish_history = []
        self._upload_history = []
        self._uploads_pending = {}
        self._sync_history = []
        self._tasks = []
        self._maintenance_report = None
        self._type_ids = self._DEFAULT_TYPE_IDS[:]
        self._seen_unit_ids = set()
        self._lock = threading.RLock()
        self._uuidgen = random.Random()
        self._uuidgen.seed(0)
        self._unitmaker = units.UnitMaker(self._seen_unit_ids)
        self._shutdown = False

    def __enter__(self):
        return self

    def __exit__(self, *_args, **_kwargs):
        self._shutdown = True

    @property
    def _all_units(self):
        # Get all units in all repos.
        #
        # Cannot be used to modify units.
        return list(self._units_by_key.values())

    def _repo_units(self, repo_id):
        # Get all units in a particular repo.
        #
        # Cannot be used to modify units.
        out = []
        for key in self._repo_unit_keys.get(repo_id) or []:
            out.append(self._units_by_key[key])
        return out

    def _remove_unit(self, repo_id, unit):
        # Remove a single unit from a repository.
        unit_key = units.make_unit_key(unit)

        # Ensure unit is no longer referenced from the repo
        self._repo_unit_keys[repo_id].remove(unit_key)

        # Ensure repo ID is no longer in memberships
        new_repos = [id for id in unit.repository_memberships if id != repo_id]
        self._units_by_key[unit_key] = attr.evolve(
            unit, repository_memberships=new_repos
        )

    def _remove_clashing_units(self, repo_id, unit):
        # In preparation for adding 'unit' into a repo, remove any existing
        # units from that repo which would clash, in a compatible manner as Pulp.
        #
        # Currently this is a special case only for File units: although their
        # unit_key consists of more than just 'name' making it technically possible
        # to have multiple files with the same name in a repo, Pulp has special logic
        # to try to prevent this, so we do the same here.
        if isinstance(unit, FileUnit):
            for existing in self._repo_units(repo_id):
                if isinstance(existing, FileUnit) and existing.path == unit.path:
                    self._remove_unit(repo_id, existing)

    def _insert_repo_units(self, repo_id, units_to_add):
        # Insert an iterable of units into a specific repo.
        #
        # If a unit with the same key exists in multiple repos, this will update
        # matching units across *all* of those repos - same as a real pulp server.

        repo_unit_keys = self._repo_unit_keys.setdefault(repo_id, set())
        memberships = []
        if repo_id is not None:
            memberships.append(repo_id)

        for unit in units_to_add:
            # Always consume a unit ID from the unitmaker even if we don't actually
            # need it. The point of this is to ensure that if you serialize/deserialize
            # a fake client, e.g. as done in pubtools-pulp, then the freshly created
            # client will not try to use the same sequence of IDs as already used in
            # the units we've just deserialized.
            unit_id = self._unitmaker.next_unit_id()

            if not unit.unit_id:
                unit = attr.evolve(unit, unit_id=unit_id)

            self._seen_unit_ids.add(unit.unit_id)

            # Unit belongs to the repo we're adding it to.
            # Note: this may be further merged with an existing unit's
            # repository_memberships a few lines below.
            unit = attr.evolve(
                unit,
                repository_memberships=(unit.repository_memberships or [])
                + memberships,
            )

            if repo_id is not None:
                # Unit might be replacing earlier units in same repo.
                self._remove_clashing_units(repo_id, unit)

            unit_key = units.make_unit_key(unit)
            old_unit = self._units_by_key.get(unit_key)
            self._units_by_key[unit_key] = units.merge_units(old_unit, unit)
            repo_unit_keys.add(unit_key)

    def _ensure_alive(self):
        if self._shutdown:
            # We are technically capable of working just fine after shutdown,
            # but the point of this class is to be an accurate stand-in for
            # a real client, so raise the same kind of exception here
            raise RuntimeError("cannot schedule new futures after shutdown")

    def search_repository(self, criteria=None):
        self._ensure_alive()

        criteria = criteria or Criteria.true()
        repos = []

        # Pass the criteria through the code used by the real client to build
        # up the Pulp query. We don't actually *use* the resulting query since
        # we're not accessing a real Pulp server. The point is to ensure the
        # same validation and error behavior as used by the real client also
        # applies to the fake.
        search_for_criteria(criteria, Repository)

        try:
            for repo in self._repositories:
                if match_object(criteria, repo):
                    repos.append(self._attach_repo(repo))
        except Exception as ex:  # pylint: disable=broad-except
            return f_return_error(ex)

        # callers should not make any assumption about the order of returned
        # values. Encourage that by returning output in unpredictable order
        random.shuffle(repos)
        return self._prepare_pages(repos)

    def search_content(self, criteria=None):
        self._ensure_alive()

        criteria = criteria or Criteria.true()
        out = []

        # Pass the criteria through the code used by the real client to build
        # up the Pulp query. We don't actually *use* the resulting query since
        # we're not accessing a real Pulp server. The point is to ensure the
        # same validation and error behavior as used by the real client also
        # applies to the fake.
        prepared_search = search_for_criteria(criteria, Unit)

        available_type_ids = set(self._type_ids)
        missing_type_ids = set(prepared_search.type_ids or []) - available_type_ids
        if missing_type_ids:
            return f_return_error(
                PulpException(
                    "following type ids are not supported on the server: %s"
                    % ",".join(missing_type_ids)
                )
            )

        for unit in self._all_units:
            if (
                prepared_search.type_ids
                and unit.content_type_id not in prepared_search.type_ids
            ):
                continue
            if match_object(criteria, unit):
                out.append(unit)

        # callers should not make any assumption about the order of returned
        # values. Encourage that by returning output in unpredictable order
        random.shuffle(out)
        return self._prepare_pages(out)

    def copy_content(self, from_repository, to_repository, criteria=None):
        self._ensure_alive()

        from_id = from_repository.id
        to_id = to_repository.id

        found = list(from_repository.search_content(criteria).result())

        # Units are being copied to this repo, so that value obviously must appear
        # in repository_memberships from now on.
        found = [attr.evolve(unit, repository_memberships=[to_id]) for unit in found]

        # Now put the found units into the destination repo.
        # Any kind of merging or replacing of units is handled within this step.
        self._insert_repo_units(to_id, found)

        # Arbitrarily limit the number of units included per task. The point is
        # to enforce that the caller doesn't expect any specific number of tasks.
        tasks = []
        while found:
            next_batch = found[:5]
            found = found[5:]
            tasks.append(
                Task(
                    id=self._next_task_id(),
                    repo_id=from_id,
                    completed=True,
                    succeeded=True,
                    units=units.with_key_only(next_batch),
                )
            )

        if not tasks:
            # This indicates that nothing was found at all.
            # That's fine, just return a task with empty units.
            tasks.append(
                Task(
                    id=self._next_task_id(),
                    repo_id=from_id,
                    completed=True,
                    succeeded=True,
                    units=[],
                )
            )

        return f_proxy(f_return(tasks))

    def update_content(self, unit):
        self._ensure_alive()

        if not unit.unit_id:
            raise ValueError("unit_id missing on call to update_content()")

        # The unit has to exist.
        existing_unit = None
        for candidate in self._all_units:
            if (
                candidate.content_type_id == unit.content_type_id
                and candidate.unit_id == unit.unit_id
            ):
                existing_unit = candidate
                break
        else:
            return f_return_error(PulpException("unit not found: %s" % unit.unit_id))

        # OK, we have a unit to update. Figure out which fields we can update.
        update = {}
        for fld in unit._usermeta_fields():
            update[fld.name] = getattr(unit, fld.name)

        updated_unit = attr.evolve(existing_unit, **update)

        unit_key = units.make_unit_key(updated_unit)
        self._units_by_key[unit_key] = updated_unit

        return f_return()

    def search_distributor(self, criteria=None):
        self._ensure_alive()

        criteria = criteria or Criteria.true()
        distributors = []

        search_for_criteria(criteria, Distributor)

        try:
            for repo in self._repositories:
                for distributor in repo.distributors:
                    if match_object(criteria, distributor):
                        distributors.append(attr.evolve(distributor, repo_id=repo.id))
        except Exception as ex:  # pylint: disable=broad-except
            return f_return_error(ex)

        random.shuffle(distributors)
        return self._prepare_pages(distributors)

    def search_task(self, criteria=None):
        self._ensure_alive()
        tasks = []

        criteria = criteria or Criteria.true()
        search_for_criteria(criteria)

        try:
            for task in self._tasks:
                if match_object(criteria, task):
                    tasks.append(task)
        except Exception as ex:  # pylint: disable=broad-except
            return f_return_error(ex)

        random.shuffle(tasks)
        return self._prepare_pages(tasks)

    def _search_repo_units(self, repo_id, criteria):
        criteria = criteria or Criteria.true()

        # Pass the criteria through the same handling as used by the real client
        # for serialization, to ensure we reject criteria also rejected by real client.
        # We don't actually use the result, this is only for validation.
        search_for_criteria(criteria, Unit)

        repo_f = self.get_repository(repo_id)
        if repo_f.exception():
            return repo_f

        repo_units = self._repo_units(repo_id)
        out = []

        try:
            for unit in repo_units:
                if match_object(criteria, unit):
                    out.append(unit)
        except Exception as ex:  # pylint: disable=broad-except
            return f_return_error(ex)

        random.shuffle(out)
        return self._prepare_pages(out)

    def _prepare_pages(self, resource_list):
        # Split resource_list into pages
        # resource_list: list of objects that paginated
        page_data = []
        current_page_data = []
        while resource_list:
            next_elem = resource_list.pop()
            current_page_data.append(next_elem)
            if len(current_page_data) == self._PAGE_SIZE and resource_list:
                page_data.append(current_page_data)
                current_page_data = []

        page_data.append(current_page_data)

        page = Page()
        next_page = None
        for batch in reversed(page_data):
            page = Page(data=batch, next=next_page)
            next_page = f_proxy(f_return(page))

        return f_proxy(f_return(page))

    def get_repository(self, repository_id):
        if not isinstance(repository_id, six.string_types):
            raise TypeError("Invalid argument: id=%s" % id)

        data = self.search_repository(Criteria.with_id(repository_id)).result().data
        if len(data) != 1:
            return f_return_error(
                PulpException("Repository id=%s not found" % repository_id)
            )

        return f_proxy(f_return(data[0]))

    def get_maintenance_report(self):
        self._ensure_alive()

        if self._maintenance_report:
            report = MaintenanceReport._from_data(json.loads(self._maintenance_report))
        else:
            report = MaintenanceReport()
        return f_proxy(f_return(report))

    def set_maintenance(self, report):
        self._ensure_alive()

        report_json = json.dumps(report._export_dict(), indent=4, sort_keys=True)
        report_fileobj = StringIO(report_json)

        repo = self.get_repository("redhat-maintenance").result()

        # upload updated report to repository and publish
        upload_ft = repo.upload_file(report_fileobj, "repos.json")

        publish_ft = f_flat_map(upload_ft, lambda _: repo.publish())
        self._maintenance_report = report_json

        return f_proxy(publish_ft)

    def get_content_type_ids(self):
        self._ensure_alive()

        return f_proxy(f_return(self._type_ids))

    def _do_upload_file(self, upload_id, file_obj):
        # We keep track of uploaded content as we may need it at import time.
        buffer = six.BytesIO()
        self._uploads_pending[upload_id] = buffer

        is_file_obj = "close" in dir(file_obj)
        if not is_file_obj:
            file_obj = open(file_obj, "rb")

        def do_next_upload(checksum, size):
            data = file_obj.read(1024 * 1024)
            if data:
                if isinstance(data, six.text_type):
                    data = data.encode("utf-8")
                buffer.write(data)
                checksum.update(data)
                size += len(data)
                return do_next_upload(checksum, size)
            return f_return((checksum.hexdigest(), size))

        out = f_flat_map(f_return(), lambda _: do_next_upload(hashlib.sha256(), 0))

        out.add_done_callback(lambda _: file_obj.close())

        return out

    def _do_unassociate(self, repo_id, criteria=None):
        repo_f = self.get_repository(repo_id)
        if repo_f.exception():
            return repo_f

        current = self._repo_unit_keys.get(repo_id, set())
        units_with_key = [
            {"key": key, "unit": self._units_by_key[key]} for key in current
        ]
        removed_units = set()
        kept_keys = set()

        # use type hint=Unit so that if type_ids are the goal here
        # then we will get back a properly prepared PulpSearch with
        # a populated type_ids field
        pulp_search = search_for_criteria(criteria, type_hint=Unit, type_ids_accum=None)

        for unit_with_key in units_with_key:
            unit = unit_with_key["unit"]
            if (
                pulp_search.type_ids is None
                or unit.content_type_id in pulp_search.type_ids
            ):
                removed_units.add(unit)
            else:
                kept_keys.add(unit_with_key["key"])

        self._repo_unit_keys[repo_id] = kept_keys

        task = Task(
            id=self._next_task_id(),
            repo_id=repo_id,
            completed=True,
            succeeded=True,
            units=units.with_key_only(removed_units),
        )

        return f_return([task])

    def _request_upload(self):
        upload_request = {
            "_href": "/pulp/api/v2/content/uploads/%s/" % self._next_request_id(),
            "upload_id": "%s" % self._next_request_id(),
        }

        return f_return(upload_request)

    def _do_import(
        self, repo_id, upload_id, unit_type_id, unit_key, unit_metadata=None
    ):
        repo_f = self.get_repository(repo_id)
        if repo_f.exception():
            # Repo can't be found, let that exception propagate
            return repo_f

        repo = repo_f.result()

        # Get the uploaded content we're about to import; though it's not
        # guaranteed to be present (e.g. erratum has no file).
        # If not present, we just use an empty BytesIO.
        upload_content = self._uploads_pending.pop(upload_id, six.BytesIO())
        upload_content.seek(0)

        new_units = self._unitmaker.make_units(
            unit_type_id, unit_key, unit_metadata, upload_content, repo_id
        )
        new_units = [
            attr.evolve(u, repository_memberships=[repo.id]) for u in new_units
        ]

        self._insert_repo_units(repo_id, new_units)

        task = Task(id=self._next_task_id(), completed=True, succeeded=True)

        # upload_history is a deprecated field, data is maintained for iso only.
        if unit_type_id == "iso":
            self._upload_history.append(
                Upload(repo, [task], unit_key["name"], unit_key["checksum"])
            )

        return f_return([task])

    def _delete_resource(self, resource_type, resource_id):
        if resource_type == "repositories":
            return self._delete_repository(resource_id)

        match = re.match(r"^repositories/([^/]+)/distributors$", resource_type)
        if match:
            return self._delete_distributor(match.group(1), resource_id)

        # There is no way to get here using public API
        raise AssertionError(
            "Asked to delete unexpected %s" % resource_type
        )  # pragma: no cover

    def _delete_repository(self, repo_id):
        with self._lock:
            found = False
            for idx, repo in enumerate(self._repositories):
                if repo.id == repo_id:
                    found = True
                    break

            if not found:
                # Deleting something which already doesn't exist is fine
                return f_return([])

            self._repositories.pop(idx)  # pylint: disable=undefined-loop-variable
            self._repo_unit_keys.pop(repo_id, None)
            return f_return(
                [Task(id=self._next_task_id(), completed=True, succeeded=True)]
            )

    def _delete_distributor(self, repo_id, distributor_id):
        with self._lock:
            repo_f = self.get_repository(repo_id)
            if repo_f.exception():
                # Repo can't be found, let that exception propagate
                return repo_f

            repo = repo_f.result()
            new_distributors = [
                dist for dist in repo.distributors if dist.id != distributor_id
            ]
            dist_found = new_distributors != repo.distributors

            if not dist_found:
                # Deleting something which already doesn't exist is fine
                return f_return([])

            idx = self._repositories.index(repo)
            self._repositories[idx] = attr.evolve(repo, distributors=new_distributors)

            return f_return(
                [
                    Task(
                        id=self._next_task_id(),
                        completed=True,
                        succeeded=True,
                        tags=[
                            "pulp:repository:%s" % repo_id,
                            "pulp:repository_distributor:%s" % distributor_id,
                            "pulp:action:remove_distributor",
                        ],
                    )
                ]
            )

    def _publish_repository(self, repo, distributors_with_config):
        repo_f = self.get_repository(repo.id)
        if repo_f.exception():
            # Repo can't be found, let that exception propagate
            return repo_f

        tasks = []
        for _ in distributors_with_config:
            tasks.append(Task(id=self._next_task_id(), completed=True, succeeded=True))

        self._publish_history.append(Publish(repo, tasks))

        return f_return(tasks)

    def _attach_repo(self, repo):
        kwargs = {}
        if repo.distributors:
            # Deep copy for accurate attach/detach semantics
            kwargs["distributors"] = [attr.evolve(dist) for dist in repo.distributors]

        repo = attr.evolve(repo, **kwargs)
        repo._set_client(self)
        return repo

    def _do_sync(self, repo_id, sync_config):  # pylint:disable = unused-argument
        repo_f = self.get_repository(repo_id)
        if repo_f.exception():
            # Repo can't be found, let that exception propagate
            return repo_f

        task = Task(id=self._next_task_id(), completed=True, succeeded=True)

        self._sync_history.append(Sync(repo_f.result(), [task], sync_config))

        return f_return([task])

    def _next_task_id(self):
        with self._lock:
            next_raw_id = self._uuidgen.randint(0, 2 ** 128)
        return str(uuid.UUID(int=next_raw_id))

    def _next_request_id(self):
        return self._next_task_id()



import datetime
import functools

import six
from frozenlist2 import frozenlist

from .attr import PULP2_PY_CONVERTER


def get_converter(field, value):
    """Given an attrs target field and an input value from Pulp,
    return a converter function which should be used to convert the Pulp value
    into a Python representation."""

    metadata_converter = field.metadata.get(PULP2_PY_CONVERTER)
    if metadata_converter:
        # explicitly defined for this field, just return it
        return metadata_converter

    # Nothing explicitly defined, but check the types, there may still be
    # some applicable default
    if field.type is datetime.datetime and isinstance(value, six.string_types):
        return read_timestamp

    return null_convert


def null_convert(value):
    return value


def read_timestamp(value):
    try:
        return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%SZ")
    except ValueError:
        # irritatingly (probably a bug), some values were missing the "Z"
        # technically meaning we don't know the timezone.
        # So we try parsing again without it and we just assume it's UTC.
        return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")


def tolerant_timestamp(value):
    # Converter for fields which can accept a timestamp string, but which
    # falls back to returning the input verbatim if conversion fails.
    #
    # Since it tolerates failed conversions, this is intended to be combined
    # with a validator.
    if isinstance(value, six.string_types):
        try:
            return read_timestamp(value)
        except ValueError:
            # Not a timestamp, conversion doesn't happen
            pass

    return value


def write_timestamp(value):
    # defaults to current time if value is None
    if value is None:
        value = datetime.datetime.utcnow()
    return value.strftime("%Y-%m-%dT%H:%M:%SZ")


def frozenlist_or_none_converter(obj, map_fn=(lambda x: x)):
    if obj is not None:
        return frozenlist(map_fn(obj))
    return None


frozenlist_or_none_sorted_converter = functools.partial(
    frozenlist_or_none_converter, map_fn=lambda x: sorted(set(x))
)



import six

from ..compat_attr import validators

instance_of = validators.instance_of

optional_str = instance_of(six.string_types + (type(None),))
optional_bool = instance_of((bool, type(None)))

# This is a workaround for the absence of deep_iterable on older attr.
# Drop it when the legacy environment is no longer required.
class OptionalListValidator(object):
    def __init__(self, member_validator):
        self.list_validator = instance_of(list)
        self.member_validator = member_validator

    def __call__(self, inst, attr, value):
        if value is None:
            # OK, no more validation
            return

        # We have a non-None value - it should be a list
        self.list_validator(inst, attr, value)

        # Validate every element
        for elem in value:
            self.member_validator(inst, attr, elem)


def optional_list_of(elem_types):
    return OptionalListValidator(instance_of(elem_types))



import datetime

from .base import Unit, unit_type

from ..attr import pulp_attrib
from ... import compat_attr as attr
from ..convert import frozenlist_or_none_sorted_converter, tolerant_timestamp
from ..validate import optional_str, instance_of


@unit_type("iso")
@attr.s(kw_only=True, frozen=True)
class FileUnit(Unit):
    """A :class:`~pubtools.pulplib.Unit` representing a generic file.

    .. versionadded:: 1.5.0
    """

    path = pulp_attrib(type=str, pulp_field="name", unit_key=True)
    """Full path for this file in the Pulp repository.

    This may include leading directory components, but
    in the common case, this is only a basename.
    """

    size = pulp_attrib(type=int, converter=int, pulp_field="size", unit_key=True)
    """Size of this file, in bytes."""

    sha256sum = pulp_attrib(
        type=str,
        pulp_field="checksum",
        converter=lambda s: s.lower() if s else s,
        unit_key=True,
    )
    """SHA256 checksum of this file, as a hex string."""

    description = pulp_attrib(
        type=str,
        pulp_field="pulp_user_metadata.description",
        default=None,
        validator=optional_str,
    )
    """A user-oriented terse description of this file.

    This field is **mutable** and may be set by
    :meth:`~FileRepository.upload_file` or
    :meth:`~Client.update_content`.

    .. versionadded:: 2.20.0
    """

    cdn_path = pulp_attrib(
        type=str,
        pulp_field="pulp_user_metadata.cdn_path",
        default=None,
        validator=optional_str,
    )
    """A path, relative to the CDN root, from which this file can be downloaded
    once published.

    This path will not point to any specific repository. However, the file must
    have been published via at least one repository before this path can be
    accessed.

    This field is **mutable** and may be set by
    :meth:`~FileRepository.upload_file` or
    :meth:`~Client.update_content`.

    .. versionadded:: 2.20.0
    """

    cdn_published = pulp_attrib(
        type=datetime.datetime,
        pulp_field="pulp_user_metadata.cdn_published",
        default=None,
        converter=tolerant_timestamp,
        validator=instance_of((datetime.datetime, type(None))),
    )
    """Approximate :class:`~datetime.datetime` in UTC at which this file first
    became available at ``cdn_path``, or ``None`` if this information is
    unavailable.

    This field is **mutable** and may be set by
    :meth:`~FileRepository.upload_file` or
    :meth:`~Client.update_content`.

    .. versionadded:: 2.20.0
    """

    content_type_id = pulp_attrib(
        default="iso", type=str, pulp_field="_content_type_id"
    )

    repository_memberships = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_sorted_converter,
        pulp_field="repository_memberships",
    )
    """IDs of repositories containing the unit, or ``None`` if this information is unavailable.

    .. versionadded:: 2.6.0
    """

    unit_id = pulp_attrib(type=str, pulp_field="_id", default=None)
    """The unique ID of this unit, if known.

    .. versionadded:: 2.20.0
    """

    @size.validator
    def _check_size(self, _, value):
        if value < 0:
            raise ValueError("Not a valid size (must be positive): %s" % value)

    @sha256sum.validator
    def _check_sha256(self, _, value):
        self._check_sum(value, "SHA256", 64)



import datetime
import re

from pubtools.pulplib._impl import compat_attr as attr
from pubtools.pulplib._impl.client.errors import PulpException
from pubtools.pulplib._impl.client.search import map_field_for_type
from pubtools.pulplib._impl.criteria import (
    TrueCriteria,
    AndCriteria,
    OrCriteria,
    FieldMatchCriteria,
    EqMatcher,
    InMatcher,
    RegexMatcher,
    ExistsMatcher,
    LessThanMatcher,
)
from pubtools.pulplib._impl.model.common import PULP2_FIELD

ABSENT = object()
CLASS_MATCHERS = []


def visit(klass):
    def wrap(func):
        CLASS_MATCHERS.append((klass, func))
        return func

    return wrap


def match_object(*args, **kwargs):  # pylint:disable=inconsistent-return-statements
    dispatch = args[0]
    for (klass, func) in CLASS_MATCHERS:
        if isinstance(dispatch, klass):
            return func(*args, **kwargs)


def get_field(field, obj):
    # Obtain a named field from a model object;
    # 'field' may be either a field name used in Pulp or a field name used
    # by our model.

    # Determine whether this field name refers to a field on the model.
    # Note that we don't care about conversion on the matcher here because:
    # - If it's a field on the model, no conversion is needed since we already
    #   are storing plain objects from the model
    # - If it's a Pulp field, conversion will be handled in pulp_value
    using_model_field = map_field_for_type(field, matcher=None, type_hint=obj.__class__)

    # Are we looking for a field on our model, or a raw Pulp field?
    if using_model_field:
        # If matching a field on the model, we can simply grab and compare
        # the attribute directly.
        return getattr(obj, field, ABSENT)

    # Otherwise, the user passed a Pulp field name (e.g. notes.eng_product_id).
    # Then we delegate to pulp_value, which can look up the corresponding model
    # field and do conversions.
    return pulp_value(field, obj)


@visit(TrueCriteria)
def match_true(*_):
    return True


@visit(AndCriteria)
def match_and(criteria, obj):
    if not criteria._operands:
        # real Pulp/mongo rejects this, so should we
        raise PulpException("Invalid AND in search query")
    for subcrit in criteria._operands:
        if not match_object(subcrit, obj):
            return False
    return True


@visit(OrCriteria)
def match_or(criteria, obj):
    if not criteria._operands:
        # real Pulp/mongo rejects this, so should we
        raise PulpException("Invalid OR in search query")
    for subcrit in criteria._operands:
        if match_object(subcrit, obj):
            return True
    return False


@visit(FieldMatchCriteria)
def match_field(criteria, obj):
    field = criteria._field
    matcher = criteria._matcher
    return match_object(matcher, field, obj)


@visit(EqMatcher)
def match_field_eq(matcher, field, obj):
    value = get_field(field, obj)
    # Pulp/mongo query is defined the same when matching
    # with a list or object/string field e.g.
    # {"foo": "bar"} will match if "foo": "bar" or
    # "foo": ["bar", "bas"]
    if isinstance(value, list) and not isinstance(matcher._value, list):
        return matcher._value in value
    return value == matcher._value


@visit(RegexMatcher)
def match_field_regex(matcher, field, obj):
    value = get_field(field, obj)
    if value is ABSENT:
        return False
    return re.search(matcher._pattern, value)


@visit(ExistsMatcher)
def match_field_exists(_matcher, field, obj):
    value = get_field(field, obj)
    return value not in [ABSENT, None]


@visit(InMatcher)
def match_in(matcher, field, obj):
    value = get_field(field, obj)
    for elem in matcher._values:
        if elem == value:
            return True
    return False


@visit(LessThanMatcher)
def match_field_less(matcher, field, obj):
    value = get_field(field, obj)
    return value < matcher._value


def pulp_value(pulp_field, obj):
    # Given a Pulp 'field' name and a PulpObject instance,
    # returns the value on the object corresponding to that Pulp field.
    # e.g. if field is 'notes.relative_url', will return obj.relative_url.
    pulp_dict = {}

    for cls_field in attr.fields(type(obj)):
        pulp_key = cls_field.metadata.get(PULP2_FIELD)
        if pulp_key:
            obj_value = getattr(obj, cls_field.name, None)
            # TODO: will we need to differentiate between "absent" and
            # "present but None"? There is no way to tell here if
            # obj_value was explicitly set to None, or if it defaulted
            # to None when the object was constructed.
            if obj_value is not None:
                pulp_dict[pulp_key] = convert_field_to_pulp(
                    obj, cls_field.name, obj_value
                )

    return pulp_dict.get(pulp_field, ABSENT)


def convert_field_to_pulp(obj, field_name, value):
    # Return a value converted from Python representation into the representation
    # expected in Pulp's API.
    # obj - some PulpObject subclass
    # field - some attr.ib name
    # value - the value possibly to convert into Pulp representation

    cls_fields = attr.fields(type(obj))
    cls_field = getattr(cls_fields, field_name)

    if cls_field.type is datetime.datetime and isinstance(value, datetime.datetime):
        # We can convert from datetime back to the ISO8601 timestamp format
        # used within Pulp
        return value.strftime("%Y-%m-%dT%H:%M:%SZ")

    # no conversion
    return value



ABSENT = object()


def lookup(value, key, default=ABSENT):
    # look up a key from a dict, supporting nested dicts.
    # This is used as a helper for mongo-like nested lookups as used for
    # certain Pulp fields, e.g. "notes.created" for looking up
    # data["notes"]["created"].
    #
    # If a default is provided, it's returned when the value is absent.
    # Otherwise, KeyError is raised, similar as for a regular dict lookup.
    keys = key.split(".")

    while value and isinstance(value, dict) and keys:
        next_key = keys.pop(0)
        value = value.get(next_key, ABSENT)

    if (value is ABSENT) or keys:
        # it's not present
        if default is ABSENT:
            raise KeyError(key)
        return default

    # it's present
    return value


def dict_put(out, key, value):
    # Put a value into a dict, supporting nested dicts in the case where
    # "key" contains dot-separated subdicts.
    #
    # Example:
    #
    #   data = {}
    #   dict_put(data, "notes.created", "today")
    #   assert data == {"notes": {"created": "today"}}
    #
    # This is the inverse of lookup, and is used while serializing model
    # objects back to Pulp dicts.
    keys = key.split(".")

    while keys:
        next_key = keys.pop(0)
        if not keys:
            # We've got the last key, so we store the value
            out[next_key] = value
        else:
            # Not the last key, so ensure there's a sub-dict.
            out = out.setdefault(next_key, {})



import pytest
import attr


def assert_model_invariants(model_object):
    """Verify that a model object (an attrs-using class) satisfies certain invariants
    which should always be true of any model object.  These include at least:

    - attrs of a defined "type" are genuinely instances of that type (or None)
    - object is hashable
    - object can be copied
    - object is equal to its copy
    - object's hash is equal to hash of its copy

    This suite of assertions is meant to protect against certain mistakes which
    can easily be made when defining an attrs-based class.
    """
    if model_object is None:
        # Test trivially passes
        return

    assert_model_fields(model_object)

    copy = attr.evolve(model_object)
    assert hash(model_object) == hash(copy), "hash differs on copied object"
    assert model_object == copy, "copied object is not equal"


def assert_model_fields(model_object):
    """Assert every defined field on a model object meets certain invariants:

    - actual value matches declared type (or None)
    - not possible to create a copy using wrong type for the field
    """
    klass = type(model_object)
    fields = attr.fields(klass)

    # Only those with a defined type are testable
    fields = [f for f in fields if f.type]

    for field in fields:
        # Every field must have a type defined
        assert field.type, "%s.%s is missing a type" % (klass.__name__, field.name)

        value = getattr(model_object, field.name)
        if value is not None:
            # These tests only valid on non-None value.

            assert isinstance(
                value, field.type
            ), "%s.%s should be a %s but is not: %s" % (
                klass.__name__,
                field.name,
                field.type.__name__,
                model_object,
            )

            if field.type is list:
                # Any lists on our model objects should be immutable.
                assert_list_immutable(value)

            # If this object is also an attrs-using class,
            # test it deeply
            if attr.has(type(value)):
                assert_model_invariants(value)

        # It should not be possible to create a copy of this object
        # with a field of an incorrect type.
        assert_type_validated(model_object, field)

        # It should not be possible to directly mutate this object.
        with pytest.raises(attr.exceptions.FrozenInstanceError):
            setattr(model_object, field.name, value)


def assert_type_validated(model_object, field):
    """Verify that model_object doesn't accept values of wrong type in field.

    This test can find some (but not all) cases where a model fails to validate
    that input data is of the expected type.
    """

    copy = None
    copy_args = {field.name: object()}

    # The way this test is written, it only works if the type is not
    # declared as `object`.  It doesn't seem like there would ever
    # be a reason to do that, but let's make sure we fail early if
    # that does happen
    assert field.type is not object, "test cannot handle type=object!"

    try:
        copy = attr.evolve(model_object, **copy_args)
    except Exception:  # pylint: disable=broad-except
        # exception is expected (we don't really care what type of exception,
        # it's unreasonably cumbersome to force e.g. TypeError everywhere)
        return

    assert copy is None, "copying with wrong type for %s was possible! copy=%s" % (
        field.name,
        copy,
    )


def assert_list_immutable(value):
    """Verifies that a given list is immutable (all operations which would normally
    write to the list instead raise).
    """

    with pytest.raises(NotImplementedError):
        value[0] = None

    with pytest.raises(NotImplementedError):
        del value[0]

    with pytest.raises(NotImplementedError):
        value += None

    with pytest.raises(NotImplementedError):
        value.insert(0, None)

    with pytest.raises(NotImplementedError):
        value.append(None)

    with pytest.raises(NotImplementedError):
        value.extend([None])

    with pytest.raises(NotImplementedError):
        value.pop()

    with pytest.raises(NotImplementedError):
        value.remove(None)



import datetime

from pubtools.pulplib._impl.model.validate import optional_list_of
from .base import Unit, unit_type, schemaless_init

from ..attr import pulp_attrib
from ... import compat_attr as attr
from ..convert import (
    frozenlist_or_none_converter,
    frozenlist_or_none_sorted_converter,
    tolerant_timestamp,
)
from ..validate import optional_str, instance_of


@attr.s(kw_only=True, frozen=True)
class RpmDependency(object):
    """
    An RPM dependency entry within :meth:`~RpmUnit.requires` and :meth:`~RpmUnit.provides`.
    """

    name = pulp_attrib(default=None, type=str, pulp_field="name")
    """
    A name of dependency - it can be name of rpm package, shared library, scriplet or 
    another name of dependency.
    """

    version = pulp_attrib(default=None, type=str, pulp_field="version")
    """
    Version of this RPM dependency.
    """

    release = pulp_attrib(default=None, type=str, pulp_field="release")
    """
    Release of this RPM dependency.
    """

    epoch = pulp_attrib(default=None, type=str, pulp_field="epoch")
    """
    Epoch of this RPM dependency.
    """

    flags = pulp_attrib(default=None, type=str, pulp_field="flags")
    """
    Flags representing relation to version of this RPM dependency.
    Can be one of GT (greater than), EQ (equal), LT (less than), GE (greater than or equal) and
    LE (less than or equal).
    """

    @classmethod
    def _from_data(cls, data):
        # Convert from raw list/dict as provided in Pulp responses into model.
        if isinstance(data, list):
            return [cls._from_data(elem) for elem in data]

        return schemaless_init(cls, data)


# Note: Pulp2 models RPM and SRPM as separate unit types,
# but there's actually no difference between their fields at all.
# This separation doesn't seem useful, so we let one class handle both.
@unit_type("rpm")
@unit_type("srpm")
@attr.s(kw_only=True, frozen=True)
class RpmUnit(Unit):
    """A :class:`~pubtools.pulplib.Unit` representing an RPM.

    .. versionadded:: 1.5.0
    """

    name = pulp_attrib(type=str, pulp_field="name", unit_key=True)
    """The name of this RPM.

    Example: the name of bash-5.0.7-1.fc30.x86_64.rpm is "bash".
    """

    version = pulp_attrib(type=str, pulp_field="version", unit_key=True)
    """The version of this RPM.

    Example: the version of bash-5.0.7-1.fc30.x86_64.rpm is "5.0.7".
    """

    release = pulp_attrib(type=str, pulp_field="release", unit_key=True)
    """The release of this RPM.

    Example: the release of bash-5.0.7-1.fc30.x86_64.rpm is "1.fc30".
    """

    arch = pulp_attrib(type=str, pulp_field="arch", unit_key=True)
    """The architecture of this RPM.

    Example: the arch of bash-5.0.7-1.fc30.x86_64.rpm is "x86_64".
    """

    epoch = pulp_attrib(default="0", type=str, pulp_field="epoch", unit_key=True)
    """The epoch of this RPM (most commonly "0").

    Example: the epoch of 3:bash-5.0.7-1.fc30.x86_64.rpm is "3".
    """

    signing_key = pulp_attrib(default=None, type=str, pulp_field="signing_key")
    """The short ID of the GPG key used to sign this RPM.

    .. seealso::
        :meth:`~pubtools.pulplib.Repository.signing_keys`
    """

    filename = pulp_attrib(default=None, type=str, pulp_field="filename")
    """Filename of this RPM.
    
    Example: mod_security_crs-3.3.0-2.el8.noarch.rpm
    """

    sourcerpm = pulp_attrib(default=None, type=str, pulp_field="sourcerpm")
    """Source rpm filename if package is not source package.

    Example: gnu-efi-3.0c-1.1.src.rpm.
    """

    # Checksums:
    # Note that "checksums" isn't part of the unit key, hence why these fields
    # are allowed to be omitted.
    md5sum = pulp_attrib(
        default=None,
        type=str,
        pulp_field="checksums.md5",
        converter=lambda s: s.lower() if s else s,
    )
    """MD5 checksum of this RPM, as a hex string."""

    sha1sum = pulp_attrib(
        default=None,
        type=str,
        pulp_field="checksums.sha1",
        converter=lambda s: s.lower() if s else s,
    )
    """SHA1 checksum of this RPM, as a hex string."""

    sha256sum = pulp_attrib(
        default=None,
        type=str,
        # Use 'checksum' field because it's indexed and therefore much faster than
        # searching for checksums.sha256.
        # It's safe since this is always stored as a copy of the sha256 checksum, see:
        # https://github.com/pulp/pulp_rpm/blob/69759d0fb9a16c0a47b1f49c78f6712e650912e1/plugins/pulp_rpm/plugins/importers/yum/upload.py#L436
        pulp_field="checksum",
        converter=lambda s: s.lower() if s else s,
        unit_key=True,
    )
    """SHA256 checksum of this RPM, as a hex string."""

    content_type_id = pulp_attrib(
        default="rpm", type=str, pulp_field="_content_type_id"
    )

    cdn_path = pulp_attrib(
        type=str,
        pulp_field="pulp_user_metadata.cdn_path",
        default=None,
        validator=optional_str,
    )
    """A path, relative to the CDN root, from which this RPM can be downloaded
    once published.

    This path will not point to any specific repository. However, the RPM must
    have been published via at least one repository before this path can be
    accessed.

    This field is **mutable** and may be set by
    :meth:`~YumRepository.upload_rpm` or
    :meth:`~Client.update_content`.

    .. versionadded:: 2.20.0
    """

    cdn_published = pulp_attrib(
        type=datetime.datetime,
        pulp_field="pulp_user_metadata.cdn_published",
        default=None,
        converter=tolerant_timestamp,
        validator=instance_of((datetime.datetime, type(None))),
    )
    """Approximate :class:`~datetime.datetime` in UTC at which this RPM first
    became available at ``cdn_path``, or ``None`` if this information is
    unavailable.

    This field is **mutable** and may be set by
    :meth:`~YumRepository.upload_rpm` or
    :meth:`~Client.update_content`.

    .. versionadded:: 2.20.0
    """

    repository_memberships = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_sorted_converter,
        pulp_field="repository_memberships",
    )
    """IDs of repositories containing the unit, or ``None`` if this information is unavailable.

    .. versionadded:: 2.6.0
    """

    unit_id = pulp_attrib(type=str, pulp_field="_id", default=None)
    """The unique ID of this unit, if known.

    .. versionadded:: 2.20.0
    """

    requires = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_converter,
        pulp_field="requires",
        validator=optional_list_of(RpmDependency),
        pulp_py_converter=RpmDependency._from_data,
    )
    """
    List of dependencies that this RPM requires or ``None`` if this information is unavailable.
    """

    provides = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_converter,
        pulp_field="provides",
        validator=optional_list_of(RpmDependency),
        pulp_py_converter=RpmDependency._from_data,
    )
    """
    List of capabilities that this RPM provides or ``None`` if this information is unavailable.
    """

    @md5sum.validator
    def _check_md5(self, _, value):
        self._check_sum(value, "MD5", 32)

    @sha1sum.validator
    def _check_sha1(self, _, value):
        self._check_sum(value, "SHA1", 40)

    @sha256sum.validator
    def _check_sha256(self, _, value):
        self._check_sum(value, "SHA256", 64)



import sys

import attr

ATTR_VERSION = tuple(int(x) for x in attr.__version__.split(".")[0:2])

# pylint: disable=invalid-name

# Wrappers around attr module to handle some Py2 vs Py3 incompatibilities.


def s(*args, **kwargs):
    # A few features we'd like to use are not available on older Python/attrs
    # versions, so we just silently disable them.
    #
    # Note this means we're technically exporting a different API on
    # Py2 vs Py3, i.e. keywords-only isn't enforced on Py2.
    # This isn't ideal, but is mitigated by the fact that any new code
    # using this library will surely not be tested solely on Py2.
    kwargs = kwargs.copy()

    if "kw_only" in kwargs and sys.version_info < (3,):  # pragma: no cover
        # This is only implemented for Python 3.
        # attrs will raise if kw_only is provided on Py2.
        del kwargs["kw_only"]

    return attr.s(*args, **kwargs)


if ATTR_VERSION < (18, 1):  # pragma: no cover

    # older attrs version didn't provide this function yet,
    # but it's easily added here.
    def fields_dict(cls):
        out = {}
        for field in attr.fields(cls):
            out[field.name] = field
        return out


else:
    fields_dict = attr.fields_dict


ib = attr.ib
Factory = attr.Factory
fields = attr.fields
evolve = attr.evolve
has = attr.has
validators = attr.validators



import collections
import re
import warnings

import six

# Due to below block:
# pylint: disable=wrong-import-position

try:
    # python 3
    Iterable = collections.abc.Iterable  # pylint: disable=invalid-name
except AttributeError:  # pragma: no cover
    # python 2
    Iterable = collections.Iterable  # pylint: disable=invalid-name

from pubtools.pulplib._impl import compat_attr as attr

from .model.unit import type_ids_for_class


class Criteria(object):
    """Represents a Pulp search criteria.

    This is an opaque class which is not intended to be created
    or used directly. Instances of this class should be obtained and
    composed by calls to the documented class methods.

    Example - searching a repository:
        .. code-block:: python

            # With Pulp 2.x / mongo, this is roughly equivalent
            # to search fragment:
            #
            #  {"notes.my-field": {"$exists": True},
            #   "notes.other-field": {"$eq": ["a", "b", "c"]}}
            #
            crit = Criteria.and_(
                Criteria.with_field('notes.my-field', Matcher.exists()),
                Criteria.with_field('notes.other-field', ["a", "b", "c"])
            )

            # criteria may now be used with client to execute a search
            repos = client.search_repository(crit)

    Example - searching across all repos for a specific content type:
        .. code-block:: python

            crit = Criteria.and_(
                Criteria.with_unit_type(RpmUnit),
                Criteria.with_field("sha256sum", Matcher.in_([
                    "49ae93732fcf8d63fe1cce759664982dbd5b23161f007dba8561862adc96d063",
                    "6b30e91df993d96df0bef0f9d232d1068fa2f7055f13650208d77b43cd7c99f6"])))

            # Will find RpmUnit instances with above sums
            units = client.search_content(crit)
    """

    exists = object()
    # exists is undocumented and deprecated, use Matcher.exists() instead

    @classmethod
    def with_id(cls, ids):
        """Args:
            ids (str, list[str])
                An id or list of ids

        Returns:
            Criteria
                criteria for finding objects matching the given ID(s)
        """
        if isinstance(ids, six.string_types):
            return cls.with_field("id", ids)
        return cls.with_field("id", Matcher.in_(ids))

    @classmethod
    def with_field(cls, field_name, field_value):
        """Args:
            field_name (str)
                The name of a field.

                Supported field names include both model fields
                and Pulp fields.  See :ref:`model_fields` for information
                about these two types of fields.

                When Pulp fields are used, field names may contain a "." to
                indicate nesting, such as ``notes.created``.

            field_value
                :class:`Matcher`
                    A matcher to be applied against the field.

                object
                    Any value, to be matched against the field via
                    :meth:`Matcher.equals`.

        Returns:
            Criteria
                criteria for finding objects where ``field_name`` is present and
                matches ``field_value``.
        """
        return FieldMatchCriteria(field_name, field_value)

    @classmethod
    def with_unit_type(cls, unit_type):
        """Args:
            unit_type (class)
                A subclass of :class:`~pubtools.pulplib.Unit`.

        Returns:
            Criteria
                criteria for finding units of type ``unit_type``.

        .. versionadded:: 2.14.0
        """

        # This is just a thin wrapper for searching on content_type_id which allows
        # the caller to avoid having to handle the (unit class <=> type id) mapping.
        type_ids = type_ids_for_class(unit_type)
        if not type_ids:
            raise TypeError("Expected a Unit type, got: %s" % repr(unit_type))

        return FieldMatchCriteria("content_type_id", Matcher.in_(type_ids))

    @classmethod
    def with_field_in(cls, field_name, field_value):
        warnings.warn(
            "with_field_in is deprecated, use Matcher.in_() instead", DeprecationWarning
        )

        return cls.with_field(field_name, Matcher.in_(field_value))

    @classmethod
    def and_(cls, *criteria):
        """Args:
            criteria (list[Criteria])
                Any number of criteria.

        Returns:
            :class:`Criteria`
                criteria for finding objects which satisfy all of the input ``criteria``.
        """
        return AndCriteria(criteria)

    @classmethod
    def or_(cls, *criteria):
        """Args:
            criteria (list[Criteria])
                Any number of criteria.

        Returns:
            Criteria
                criteria for finding objects which satisfy any of the input ``criteria``.
        """
        return OrCriteria(criteria)

    @classmethod
    def true(cls):
        """
        Returns:
            Criteria
                a criteria which always matches any object.
        """
        return TrueCriteria()


class Matcher(object):
    """Methods for matching fields within a Pulp search query.

    Instances of this class are created by the documented class methods,
    and should be used in conjunction with :class:`Criteria` methods, such
    as :meth:`Criteria.with_field`.

    .. versionadded:: 1.1.0
    """

    @classmethod
    def equals(cls, value):
        """
        Matcher for a field which must equal exactly the given value.

        Arguments:
            value (object)
                An object to match against a field.
        """
        return EqMatcher(value)

    @classmethod
    def regex(cls, pattern):
        """
        Matcher for a field which must be a string and must match the given
        regular expression.

        Arguments:
            pattern (str)
                A regular expression to match against the field.
                The expression is not implicitly anchored.

                .. warning::

                    It is not defined which specific regular expression syntax is
                    supported. For portable code, callers are recommended to use
                    only the common subset of PCRE-compatible and Python-compatible
                    regular expressions.

        Raises:
            :class:`re.error`
                If the given pattern is not a valid regular expression.

        Example:

            .. code-block:: python

                # Would match any Repository where notes.my-field starts
                # with "abc"
                crit = Criteria.with_field('notes.my-field', Matcher.regex("^abc"))
        """

        return RegexMatcher(pattern)

    @classmethod
    def exists(cls):
        """
        Matcher for a field which must exist, with no specific value.

        Example:

            .. code-block:: python

                # Would match any Repository where notes.my-field exists
                crit = Criteria.with_field('notes.my-field', Matcher.exists())
        """
        return ExistsMatcher()

    @classmethod
    def in_(cls, values):
        """
        Returns a matcher for a field whose value equals one of the specified
        input values.

        Arguments:
            values (iterable)
                An iterable of values used to match a field.

        Example:

            .. code-block:: python

                # Would match any Repository where notes.my-field is "a", "b" or "c"
                crit = Criteria.with_field(
                    'notes.my-field',
                    Matcher.in_(["a", "b", "c"])
                )
        """
        return InMatcher(values)

    @classmethod
    def less_than(cls, value):
        """
        Returns a matcher for a field whose value is less than the specified input
        value.

        Arguments:
            value (object)
                An object to match against the field

        Example:
            .. code-block:: python

                # would match where last_publish is before "2019-08-27T00:00:00Z"
                # date comparison requires a datetime.datetime object
                crit = Criteria.with_field(
                    'last_publish',
                    Matcher.less_than(datetime.datetime(2019, 8, 27, 0, 0, 0))
                )

        .. versionadded:: 2.1.0
        """
        return LessThanMatcher(value)

    def _map(self, _fn):
        # Internal-only: return self with matched value mapped through
        # the given function. Intended to be overridden in subclasses
        # to support field conversions between Pulp and Python.
        return self


@attr.s
class RegexMatcher(Matcher):
    _pattern = attr.ib()

    @_pattern.validator
    def _check_pattern(self, _, pattern):
        # It must be a string.
        # Need an explicit check here because re.compile also succeeds
        # on already-compiled regex objects.
        if not isinstance(pattern, six.string_types):
            raise TypeError("Regex matcher expected string, got: %s" % repr(pattern))

        # Verify that the given value can really be compiled as a regex.
        re.compile(pattern)

    # Note: regex matcher does not implement _map since regex is defined only
    # in terms of strings, there are no meaningful conversions.

    def __str__(self):
        return "=~/%s/" % self._pattern


@attr.s
class EqMatcher(Matcher):
    _value = attr.ib()

    def _map(self, fn):
        return attr.evolve(self, value=fn(self._value))

    def __str__(self):
        return "==%s" % repr(self._value)


@attr.s
class InMatcher(Matcher):
    _values = attr.ib()

    @_values.validator
    def _check_values(self, _, values):
        if isinstance(values, Iterable) and not isinstance(values, six.string_types):
            return
        raise ValueError("Must be an iterable: %s" % repr(values))

    def _map(self, fn):
        return attr.evolve(self, values=[fn(x) for x in self._values])

    def __str__(self):
        return " IN %s" % repr(self._values)


@attr.s
class ExistsMatcher(Matcher):
    def __str__(self):
        return " EXISTS"


@attr.s
class LessThanMatcher(Matcher):
    _value = attr.ib()

    def _map(self, fn):
        return attr.evolve(self, value=fn(self._value))

    def __str__(self):
        return "<%s" % repr(self._value)


def coerce_to_matcher(value):
    if isinstance(value, Matcher):
        return value

    if value is Criteria.exists:
        warnings.warn(
            "Criteria.exists is deprecated, use Matcher.exists() instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return ExistsMatcher()

    return EqMatcher(value)


@attr.s
class FieldMatchCriteria(Criteria):
    _field = attr.ib()
    _matcher = attr.ib(converter=coerce_to_matcher)

    def __str__(self):
        matcher = str(self._matcher)
        out = "%s%s" % (self._field, matcher)

        if " " in matcher:
            out = "(%s)" % out
        return out


@attr.s
class AndCriteria(Criteria):
    _operands = attr.ib()

    def __str__(self):
        if not self._operands:
            return "<empty AND>"

        if len(self._operands) == 1:
            return str(self._operands[0])

        return "(" + " AND ".join([str(o) for o in self._operands]) + ")"


@attr.s
class OrCriteria(Criteria):
    _operands = attr.ib()

    def __str__(self):
        if not self._operands:
            return "<empty OR>"

        if len(self._operands) == 1:
            return str(self._operands[0])

        return "(" + " OR ".join([str(o) for o in self._operands]) + ")"


class TrueCriteria(Criteria):
    def __str__(self):
        return "TRUE"



import sys

from pubtools.pluggy import hookspec, pm


# pylint: disable=unused-argument


@hookspec
def pulp_repository_pre_publish(repository, options):
    """Invoked as the first step in publishing a Pulp repository.

    If a hookimpl returns a non-None value, that value will be used to replace
    the options for this publish. This can be used to adjust publish options
    from within a hook.

    Args:
        repository (:class:`~pubtools.pulplib.Repository`):
            The repository to be published.

        options (:class:`~pubtools.pulplib.PublishOptions`):
            The options to use in publishing.

    Returns:
        options (:class:`~pubtools.pulplib.PublishOptions`):
            The potentially adjusted options used for this publish.
    """


@hookspec
def pulp_repository_published(repository, options):
    """Invoked after a Pulp repository has been successfully published.

    Args:
        repository (:class:`~pubtools.pulplib.Repository`):
            The repository which has been published.

        options (:class:`~pubtools.pulplib.PublishOptions`):
            The options used for this publish.
    """


pm.add_hookspecs(sys.modules[__name__])



import warnings

from .client import FakeClient


class FakeController(object):
    """A controller for a fake :class:`~pubtools.pulplib.Client`, to be used
    within automated tests.

    This class provides a client which has the same public interface as
    :class:`~pubtools.pulplib.Client`. This client can do most of the same
    actions as a real client, but uses a simple in-memory implementation rather
    than issuing requests to a remote Pulp server. The client's data may be
    inspected and modified via the FakeController instance.

    Example:

        .. code-block:: python

            # imagine we have a function like this
            def delete_old_repos(client, days):
                # code here using client to find/delete all repos older
                # than given number of days
                ...

            # then we can use a fake client to test it like this
            def test_delete_old_repos():
                controller = FakeController()

                # Insert repositories into fake data
                repo_now = YumRepository(id='current-repo', created=datetime.utcnow())
                repo_old = FileRepository(id='old-repo', created=datetime.min)

                controller.insert_repository(repo_now)
                controller.insert_repository(repo_old)

                # Run the code under test against the fake client

                # It should succeed without crashing
                delete_old_repos(controller.client, 5)

                # The set of repos should now contain only the recent repo
                assert controller.repositories == [repo_now]

    **Limitations:**

        While :class:`~pubtools.pulplib.Client` allows searching on any fields
        which exist within Pulp's database, the fake client only supports searching
        on fields known to the :ref:`schemas` used by this library.
    """

    def __init__(self):
        self.client = FakeClient()
        """The client instance attached to this controller."""

    @property
    def repositories(self):
        """The list of existing repositories in the fake client."""
        return self.client._repositories[:]

    def insert_repository(self, repository):
        """Add a repository to the set of existing repositories in the fake client.

        Args:
            repository (:class:`~pubtools.pulplib.Repository`)
                A repository object to insert.
        """
        self.client._repositories.append(repository)

    def insert_units(self, repository, units):
        """Add units to the set of existing content for a repository.

        Args:
            repository (:class:`~pubtools.pulplib.Repository`)
                A repository object, or ``None`` to insert units as orphans.
            units (list[:class:`~pubtools.pulplib.Unit`])
                A list of units to be inserted.

        .. versionadded:: 1.5.0

        .. versionadded:: 2.21.0
            Added ability to insert orphan units.
        """
        repo_id = repository.id if repository else None
        self.client._insert_repo_units(repo_id, units)

    @property
    def content_type_ids(self):
        """The list of content type IDs the fake client will claim to support.

        .. versionadded:: 1.4.0
        """
        return self.client._type_ids[:]

    def set_content_type_ids(self, type_ids):
        """Set the list of content type IDs the fake client will claim to support.

        Args:
            type_ids (list[str])
                A list of content type IDs (e.g. "rpm", "erratum", ...)

        .. versionadded:: 1.4.0
        """
        self.client._type_ids = type_ids[:]

    @property
    def publish_history(self):
        """A list of repository publishes triggered via this client.

        Each element of this list is a named tuple with the following attributes,
        in order:

            ``repository``:
                :class:`~pubtools.pulplib.Repository` for which publish was triggered
            ``tasks``:
                list of :class:`~pubtools.pulplib.Task` generated as a result
                of this publish
        """
        return self.client._publish_history[:]

    @property
    def sync_history(self):
        """A list of repository syncs triggered via this client.

        Each element of this list is a named tuple with the following attributes,
        in order:

            ``repository``:
                :class:`~pubtools.pulplib.Repository` for which sync was triggered
            ``tasks``:
                list of :class:`~pubtools.pulplib.Task` generated as a result
                of this sync
            ``sync_config``:
                :class:`~pubtools.pulplib.SyncConfig` (of the appropriate subclass) used for this sync

        .. versionadded:: 2.5.0
        """
        return self.client._sync_history[:]

    @property
    def upload_history(self):
        # A list of upload tasks triggered via this client.
        #
        # Each element of this list is a named tuple with the following attributes,
        # in order:
        #
        #     ``repository``:
        #         :class:`~pubtools.pulplib.Repository` for which upload was triggered
        #     ``tasks``:
        #         list of :class:`~pubtools.pulplib.Task` generated as a result
        #         of this upload
        #     ``name`` (str):
        #         the remote path used
        #     ``sha256`` (str):
        #         checksum of the file uploaded
        #
        # Deprecated: structure was unintentionally specific to ISO units and is also
        # unnecessary as uploading will make unit(s) available in the relevant repo.
        #
        warnings.warn(
            "upload_history is deprecated, check repo units instead", DeprecationWarning
        )

        return self.client._upload_history[:]

    @property
    def tasks(self):
        """The list of existing tasks in the fake client."""
        return self.client._tasks[:]

    def insert_task(self, task):
        """Add a task to the set of existing tasks in the fake client.

        Args:
            task (:class:`~pubtools.pulplib.Task`)
                A task object to insert.
        """
        self.client._tasks.append(task)



class PulpException(Exception):
    """Raised when the Pulp server has responded with an unrecoverable error
    (generally a failed HTTP response which persisted over several retries).
    """


class TaskFailedException(PulpException):
    """Raised when a Pulp task has completed with errors."""

    def __init__(self, task):
        self.task = task
        """The :class:`~pubtools.pulplib.Task` which has failed.
        Error details may be accessed from the task.
        """

        super(TaskFailedException, self).__init__("Task %s failed" % task.id)


class MissingTaskException(PulpException):
    # this is not public API, this exception should be extremely rare
    pass



import datetime

import attr
import sys
import pytest

import pubtools.pulplib

from .assertions import assert_model_invariants


def test_invariants(model_object):
    """Test that the given object satisfies all expected invariants of model objects."""

    assert_model_invariants(model_object)


@pytest.mark.parametrize("field_name", ["repository_memberships"])
def test_stable_order(model_object, field_name):
    """Test that certain fields on the given object have a stable ordering applied."""

    if not hasattr(model_object, field_name):
        pytest.skip("This object does not have %s" % field_name)

    updates1 = {field_name: ["c", "a", "b"]}
    updates2 = {field_name: ["a", "c", "b"]}

    # Request two different updates on the object.
    obj1 = attr.evolve(model_object, **updates1)
    obj2 = attr.evolve(model_object, **updates2)

    # The result should be exactly the same in both cases.
    assert getattr(obj1, field_name) == getattr(obj2, field_name)


@pytest.mark.parametrize("field_name", ["repository_memberships"])
def test_unique(model_object, field_name):
    """Test that certain list fields on the given object enforce uniqueness."""

    if not hasattr(model_object, field_name):
        pytest.skip("This object does not have %s" % field_name)

    updates1 = {field_name: ["a", "a", "b", "b", "c"]}
    updates2 = {field_name: ["a", "b", "c"]}

    # Request two different updates on the object.
    obj1 = attr.evolve(model_object, **updates1)
    obj2 = attr.evolve(model_object, **updates2)

    # The result should be exactly the same in both cases.
    assert getattr(obj1, field_name) == getattr(obj2, field_name)


@pytest.mark.parametrize("field_name", ["cdn_published"])
def test_datetime_str(model_object, field_name):
    """Test that certain datetime fields on the given object accept timestamps."""

    if not hasattr(model_object, field_name):
        pytest.skip("This object does not have %s" % field_name)

    updates1 = {field_name: "2021-12-14T09:59:00Z"}
    updates2 = {field_name: datetime.datetime(2021, 12, 14, 9, 59, 0)}

    # Request two different updates on the object.
    obj1 = attr.evolve(model_object, **updates1)
    obj2 = attr.evolve(model_object, **updates2)

    # The result should be exactly the same in both cases.
    assert getattr(obj1, field_name) == getattr(obj2, field_name)

    # However if I try to set the same field to an *invalid* string, I should
    # get the usual TypeError from trying to store a str in a datetime field.
    with pytest.raises(TypeError):
        attr.evolve(model_object, **{field_name: "oops not a timestamp"})


def public_model_objects():
    """Returns a default-constructed instance of every public model class
    found in pubtools.pulplib.
    """
    public_symbols = dir(pubtools.pulplib)
    public_symbols = [sym for sym in public_symbols if not sym.startswith("_")]

    public_classes = [getattr(pubtools.pulplib, sym) for sym in public_symbols]
    public_classes = [klass for klass in public_classes if attr.has(klass)]

    return [default_object(klass) for klass in public_classes]


def default_for_field(field):
    # Certain fields with known strict semantics
    value_by_name = {"sha256sum": "a" * 64, "sha1sum": "b" * 40, "md5sum": "c" * 32}

    # Generic defaults for certain types
    value_by_type = {
        str: "test",
        int: 123,
        datetime.datetime: datetime.datetime(2019, 9, 4, 12, 9, 30),
    }

    if sys.version_info.major == 2:
        value_by_type[basestring] = "test"

    if field.name in value_by_name:
        return value_by_name[field.name]
    if field.type in value_by_type:
        return value_by_type[field.type]

    # Try a default-constructed object.
    return field.type()


def all_fields(klass):
    """Returns all attr fields on klass and all base classes."""
    out = attr.fields(klass)
    for base in klass.__bases__:
        if attr.has(base):
            out = out + attr.fields(base)
    return out


def default_object(klass):
    """Return an instance of attrs-using klass, constructed with minimal arguments.
    This means:

    - any fields with a declared default will use that default
    - any other fields have basic defaults injected
    """
    args = {}

    for field in all_fields(klass):
        if field.default is not attr.NOTHING:
            # There's a default, then no arg needed
            continue
        if field.name.startswith("_"):
            # Internal field
            continue
        if field.name in args:
            # Already calculated field from a derived class,
            # no need to recalculate for base
            continue
        if not field.type:
            # No type declared, try None as default
            args[field.name] = None
        else:
            args[field.name] = default_for_field(field)

    return klass(**args)


@pytest.fixture(params=public_model_objects(), ids=lambda object: type(object).__name__)
def model_object(request):
    """This fixture yields an object for every public attrs-using class
    in the library."""
    yield request.param



import re

from .base import Unit, unit_type, schemaless_init

from ..attr import pulp_attrib
from ... import compat_attr as attr
from ..convert import frozenlist_or_none_converter, frozenlist_or_none_sorted_converter
from ..validate import optional_list_of


@attr.s(kw_only=True, frozen=True)
class ModulemdDependency(object):
    """
    A module dependency entry within :meth:`~ModulemdUnit.dependencies`.
    """

    name = pulp_attrib(type=str, default=None)
    """
    The name of this dependency.
    """
    stream = pulp_attrib(type=str, default=None)
    """
    The stream of this dependency.
    """

    @classmethod
    def _from_data(cls, data):
        # Convert from raw list/dict as provided in Pulp responses into model.
        if isinstance(data, list):
            raw_data = []
            for item in data:
                for key, value in item.items():
                    # skip key == "platform", not usable as dependency
                    if key != "platform":
                        # value list might be empty
                        if value:
                            for stream in value:
                                raw_data.append({"name": key, "stream": stream})
                        else:
                            raw_data.append({"name": key})

            return [cls._from_data(elem) for elem in raw_data]

        return schemaless_init(cls, data)


@unit_type("modulemd")
@attr.s(kw_only=True, frozen=True)
class ModulemdUnit(Unit):
    """A :class:`~pubtools.pulplib.Unit` representing a modulemd document.

    .. versionadded:: 1.5.0
    """

    name = pulp_attrib(type=str, pulp_field="name", unit_key=True)
    """The name of this module.

    Example: the name of javapackages-tools:201801:20180813043155:dca7b4a4:aarch64
    is "javapackages-tools".
    """

    stream = pulp_attrib(type=str, pulp_field="stream", unit_key=True)
    """The stream of this module.

    Example: the stream of javapackages-tools:201801:20180813043155:dca7b4a4:aarch64
    is "201801".
    """

    version = pulp_attrib(type=int, pulp_field="version", converter=int, unit_key=True)
    """The version of this module.

    Example: the version of javapackages-tools:201801:20180813043155:dca7b4a4:aarch64
    is 20180813043155.
    """

    context = pulp_attrib(type=str, pulp_field="context", unit_key=True)
    """The context of this module.

    Example: the context of javapackages-tools:201801:20180813043155:dca7b4a4:aarch64
    is "dca7b4a4".
    """

    arch = pulp_attrib(type=str, pulp_field="arch", unit_key=True)
    """The architecture of this module.

    Example: the arch of javapackages-tools:201801:20180813043155:dca7b4a4:aarch64
    is "aarch64.
    """

    content_type_id = pulp_attrib(
        default="modulemd", type=str, pulp_field="_content_type_id"
    )

    repository_memberships = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_sorted_converter,
        pulp_field="repository_memberships",
    )
    """IDs of repositories containing the unit, or ``None`` if this information is unavailable.

    .. versionadded:: 2.6.0
    """

    unit_id = pulp_attrib(type=str, pulp_field="_id", default=None)
    """The unique ID of this unit, if known.

    .. versionadded:: 2.20.0
    """

    artifacts = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_converter,
        pulp_field="artifacts",
    )
    """List of artifacts included in the module.

    Typically a list of RPM NEVRAs (no '.rpm' extension) including binary, debug
    and source RPMs.

    Example:

    .. code-block:: python

        ["perl-version-7:0.99.24-441.module+el8.3.0+6718+7f269185.src",
         "perl-version-7:0.99.24-441.module+el8.3.0+6718+7f269185.x86_64"]

    """

    profiles = pulp_attrib(type=dict, pulp_field="profiles", default=None)
    """The profiles of this modulemd unit."""

    dependencies = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_converter,
        pulp_field="dependencies",
        validator=optional_list_of(ModulemdDependency),
        pulp_py_converter=ModulemdDependency._from_data,
    )
    """
    List of dependencies that this modulemd requires for runtime or ``None`` if this information is unavailable.
    """

    @property
    def artifacts_filenames(self):
        """
        RPM filenames for artifacts in this module (as opposed to the RPM
        NEVRAs returned by :meth:`artifacts`).

        Returns:
            List[str]
                Artifact RPM filenames.
        """
        regex = r"\d+:"
        reg = re.compile(regex)

        out = set()
        for rpm_nevra in self.artifacts or []:
            out.add(reg.sub("", rpm_nevra, count=1) + ".rpm")
        return out

    @property
    def nsvca(self):
        """
        Returns nsvca string of this module.

        Example: "virt:av:8040020210622174547:522a0ee4:arch"
        """
        return ":".join(
            (self.name, self.stream, str(self.version), self.context, self.arch)
        )



import six

from .base import Unit, PulpObject, unit_type, schemaless_init

from ..attr import pulp_attrib
from ... import compat_attr as attr
from ..validate import optional_bool, optional_list_of, optional_str, instance_of
from ..convert import frozenlist_or_none_sorted_converter, frozenlist_or_none_converter


@attr.s(kw_only=True, frozen=True)
class ErratumReference(PulpObject):
    """A reference within a :meth:`~ErratumUnit.references` list."""

    href = pulp_attrib(
        type=str, pulp_field="href", default=None, validator=optional_str
    )
    """A URL."""

    id = pulp_attrib(type=str, pulp_field="id", default=None, validator=optional_str)
    """A short ID for the reference, unique within this erratum."""

    title = pulp_attrib(
        type=str, pulp_field="title", default=None, validator=optional_str
    )
    """A title for the reference; analogous to the 'title' attribute
    in HTML.
    """

    type = pulp_attrib(
        type=str, pulp_field="type", default=None, validator=optional_str
    )
    """Type of reference. This defines the expected target of the URL
    and includes at least:

    - "self": reference to a page for this advisory
    - "bugzilla": reference to a bug
    - "other": any other kind of reference
    """

    @classmethod
    def _from_data(cls, data):
        if data is None:
            return None

        # Convert from raw list/dict in Pulp response
        if isinstance(data, list):
            return [cls._from_data(elem) for elem in data]

        return schemaless_init(cls, data)


@attr.s(kw_only=True, frozen=True)
class ErratumModule(PulpObject):
    """A module entry within a :meth:`~ErratumUnit.pkglist`."""

    name = pulp_attrib(
        type=str, pulp_field="name", default=None, validator=optional_str
    )
    """Module name."""

    stream = pulp_attrib(
        type=str, pulp_field="stream", default=None, validator=optional_str
    )
    """Module stream."""

    version = pulp_attrib(
        type=str, pulp_field="version", default=None, validator=optional_str
    )
    """Module version."""

    context = pulp_attrib(
        type=str, pulp_field="context", default=None, validator=optional_str
    )
    """Module context."""

    arch = pulp_attrib(
        type=str, pulp_field="arch", default=None, validator=optional_str
    )
    """Module architecture."""

    @classmethod
    def _from_data(cls, data):
        return schemaless_init(cls, data)


@attr.s(kw_only=True, frozen=True)
class ErratumPackage(PulpObject):
    """A package (RPM) entry within a :meth:`~ErratumUnit.pkglist`."""

    arch = pulp_attrib(
        type=str, pulp_field="arch", default=None, validator=optional_str
    )
    """RPM architecture."""

    filename = pulp_attrib(
        type=str, pulp_field="filename", default=None, validator=optional_str
    )
    """RPM filename (basename)."""

    epoch = pulp_attrib(
        type=str, pulp_field="epoch", default=None, validator=optional_str
    )
    """RPM epoch."""

    name = pulp_attrib(
        type=str, pulp_field="name", default=None, validator=optional_str
    )
    """RPM name (e.g. "bash-4.0.1-1.el7.x86_64.rpm" name is "bash")"""

    version = pulp_attrib(
        type=str, pulp_field="version", default=None, validator=optional_str
    )
    """RPM version (e.g. "bash-4.0.1-1.el7.x86_64.rpm" version is "4.0.1")"""

    release = pulp_attrib(
        type=str, pulp_field="release", default=None, validator=optional_str
    )
    """RPM release (e.g. "bash-4.0.1-1.el7.x86_64.rpm" version is "1.el7")"""

    src = pulp_attrib(type=str, pulp_field="src", default=None, validator=optional_str)
    """Filename of the source RPM from which this RPM was built; equal to
    :meth:`filename` for the source RPM itself.
    """

    reboot_suggested = pulp_attrib(
        type=bool, pulp_field="reboot_suggested", default=None, validator=optional_bool
    )
    """True if rebooting host machine is recommended after installing this package."""

    md5sum = attr.ib(type=str, default=None, validator=optional_str)
    """MD5 checksum of this RPM in hex string form, if available."""

    sha1sum = attr.ib(type=str, default=None, validator=optional_str)
    """SHA1 checksum of this RPM in hex string form, if available."""

    sha256sum = attr.ib(type=str, default=None, validator=optional_str)
    """SHA256 checksum of this RPM in hex string form, if available."""

    @classmethod
    def _from_data(cls, data):
        if isinstance(data, list):
            return [cls._from_data(elem) for elem in data]

        data_updated = data.copy()

        # parse the odd 'sum' structure, which is a list of form:
        # [<algo>, <hexdigest>, <algo>, <hexdigest>, ...]
        sums = {}
        raw_sum = data_updated.pop("sum", [])
        while raw_sum:
            sums[raw_sum[0] + "sum"] = raw_sum[1]
            raw_sum = raw_sum[2:]
        data_updated.update(sums)

        return schemaless_init(cls, data_updated)

    def _to_data(self):
        # Handle model-to-pulp special cases for a couple of fields.
        out = super(ErratumPackage, self)._to_data()

        # If reboot_suggested has no value, it's critical that we
        # omit it entirely. At least some versions of yum treat the presence
        # of this field as if the value is True, without actually looking at
        # the value, and this logic has been carried over to Pulp also.
        if "reboot_suggested" in out and out["reboot_suggested"] is None:
            del out["reboot_suggested"]

        # 'sum' is assembled from multiple fields
        sumlist = []
        if self.md5sum:
            sumlist.extend(["md5", self.md5sum])
        if self.sha1sum:
            sumlist.extend(["sha1", self.sha1sum])
        if self.sha256sum:
            sumlist.extend(["sha256", self.sha256sum])
        out["sum"] = sumlist

        return out


@attr.s(kw_only=True, frozen=True)
class ErratumPackageCollection(PulpObject):
    """A collection of packages found within an :meth:`~ErratumUnit.pkglist`.

    A non-modular advisory typically contains only a single collection, while modular
    advisories typically contain one collection per module.
    """

    name = pulp_attrib(
        type=str, pulp_field="name", default=None, validator=optional_str
    )
    """A name for this collection. The collection name has no specific meaning,
    but must be unique within an advisory.
    """

    packages = pulp_attrib(
        type=list,
        pulp_field="packages",
        default=None,
        converter=frozenlist_or_none_converter,
        validator=optional_list_of(ErratumPackage),
    )
    """List of packages within this collection.

    :type: list[ErratumPackage]
    """

    short = pulp_attrib(
        type=str, pulp_field="short", default=None, validator=optional_str
    )
    """An alternative name for this collection. In practice, this field
    is typically blank.
    """

    module = pulp_attrib(
        type=ErratumModule,
        pulp_field="module",
        default=None,
        validator=instance_of((ErratumModule, type(None))),
    )
    """An :class:`~ErratumModule` defining the module this entry is associated
    with, if any.
    """

    @classmethod
    def _from_data(cls, data):
        if data is None:
            return None

        # Convert from raw list/dict as provided in Pulp responses into model.
        if isinstance(data, list):
            return [cls._from_data(elem) for elem in data]

        data_updated = data.copy()

        if data.get("packages") is not None:
            data_updated["packages"] = ErratumPackage._from_data(data["packages"])
        if data.get("module") is not None:
            data_updated["module"] = ErratumModule._from_data(data["module"])

        return schemaless_init(cls, data_updated)


@unit_type("erratum")
@attr.s(kw_only=True, frozen=True)
class ErratumUnit(Unit):
    """A :class:`~pubtools.pulplib.Unit` representing an erratum/advisory.

    .. versionadded:: 2.17.0
    """

    id = pulp_attrib(type=str, pulp_field="id", unit_key=True, validator=optional_str)
    """The ID of this advisory.

    Example: ``"RHSA-2021:0672"``"""

    version = pulp_attrib(
        type=str, pulp_field="version", default=None, validator=optional_str
    )
    """The version of this advisory.

    Though stored as a string, this field typically takes the form of an
    integer starting at "1" and incremented whenever the advisory is modified.
    """

    status = pulp_attrib(
        type=str, pulp_field="status", default=None, validator=optional_str
    )
    """Status, typically 'final'."""

    updated = pulp_attrib(
        type=str, pulp_field="updated", default=None, validator=optional_str
    )
    """Timestamp of the last update to this advisory.

    Typically of the form '2019-12-31 06:54:41 UTC', but this is not enforced."""

    issued = pulp_attrib(
        type=str, pulp_field="issued", default=None, validator=optional_str
    )
    """Timestamp of the initial release of this advisory.

    Uses the same format as :meth:`updated`."""

    description = pulp_attrib(
        type=str, pulp_field="description", default=None, validator=optional_str
    )
    """Full human-readable description of the advisory, usually multiple lines."""

    pushcount = pulp_attrib(
        type=str, pulp_field="pushcount", default=None, validator=optional_str
    )
    """Number of times advisory has been revised and published (starting at '1')."""

    reboot_suggested = pulp_attrib(
        type=bool, pulp_field="reboot_suggested", default=None, validator=optional_bool
    )
    """True if rebooting host machine is recommended after installing this advisory."""

    from_ = pulp_attrib(
        type=str, pulp_field="from", default=None, validator=optional_str
    )
    """Contact email address for the owner of the advisory.

    Note that the canonical name for this attribute is ``from``.
    As this clashes with a Python keyword, in most contexts the attribute is
    available as an alias, ``from_``.
    """

    rights = pulp_attrib(
        type=str, pulp_field="rights", default=None, validator=optional_str
    )
    """Copyright message."""

    title = pulp_attrib(
        type=str, pulp_field="title", default=None, validator=optional_str
    )
    """Title of the advisory (e.g. 'bash bugfix and enhancement')."""

    severity = pulp_attrib(
        type=str, pulp_field="severity", default=None, validator=optional_str
    )
    """Severity of the advisory, e.g. "low", "moderate", "important" or "critical"."""

    release = pulp_attrib(
        type=str, pulp_field="release", default=None, validator=optional_str
    )
    """Release number. Typically an integer-string, initially "0"."""

    type = pulp_attrib(
        type=str, pulp_field="type", default=None, validator=optional_str
    )
    """"bugfix", "security" or "enhancement"."""

    solution = pulp_attrib(
        type=str, pulp_field="solution", default=None, validator=optional_str
    )
    """Text explaining how to apply the advisory."""

    summary = pulp_attrib(
        type=str, pulp_field="summary", default=None, validator=optional_str
    )
    """Typically a single sentence briefly describing the advisory."""

    content_types = pulp_attrib(
        type=list,
        pulp_field="pulp_user_metadata.content_types",
        converter=frozenlist_or_none_converter,
        default=None,
        validator=optional_list_of(six.string_types),
    )
    """A list of content types associated with the advisory.

    For example, "rpm" may be found in this list if the advisory has any
    associated RPMs.
    """

    references = pulp_attrib(
        type=list,
        pulp_field="references",
        pulp_py_converter=ErratumReference._from_data,
        converter=frozenlist_or_none_converter,
        default=None,
        validator=optional_list_of(ErratumReference),
    )
    """A list of references associated with the advisory."""

    pkglist = pulp_attrib(
        type=list,
        pulp_field="pkglist",
        pulp_py_converter=ErratumPackageCollection._from_data,
        converter=frozenlist_or_none_converter,
        default=None,
        validator=optional_list_of(ErratumPackageCollection),
    )
    """A list of package collections associated with the advisory."""

    content_type_id = pulp_attrib(
        default="erratum", type=str, pulp_field="_content_type_id"
    )

    repository_memberships = pulp_attrib(
        default=None,
        type=list,
        converter=frozenlist_or_none_sorted_converter,
        pulp_field="repository_memberships",
    )
    """IDs of repositories containing the unit, or ``None`` if this information is unavailable.
    """

    unit_id = pulp_attrib(type=str, pulp_field="_id", default=None)
    """The unique ID of this unit, if known.

    .. versionadded:: 2.20.0
    """



import logging
import datetime
import contextlib
from pubtools.pulplib._impl.criteria import (
    AndCriteria,
    OrCriteria,
    FieldMatchCriteria,
    TrueCriteria,
    RegexMatcher,
    EqMatcher,
    InMatcher,
    ExistsMatcher,
    LessThanMatcher,
)

from pubtools.pulplib._impl import compat_attr as attr
from pubtools.pulplib._impl.model.attr import PULP2_FIELD, PY_PULP2_CONVERTER
from pubtools.pulplib._impl.model.unit.base import Unit

LOG = logging.getLogger("pubtools.pulplib")


def all_subclasses(klass):
    out = set()
    out.add(klass)
    for subclass in klass.__subclasses__():
        out.update(all_subclasses(subclass))
    return out


def to_mongo_json(value):
    # Return a value converted to the format expected for a mongo JSON
    # expression. Only a handful of special types need explicit conversions.
    if isinstance(value, datetime.datetime):
        return {"$date": value.strftime("%Y-%m-%dT%H:%M:%SZ")}

    if isinstance(value, (list, tuple)):
        return [to_mongo_json(elem) for elem in value]

    if isinstance(value, dict):
        out = {}
        for (key, val) in value.items():
            out[key] = to_mongo_json(val)
        return out

    return value


def map_field_for_type(field_name, matcher, type_hint):
    if not type_hint:
        return None

    attrs_classes = all_subclasses(type_hint)
    attrs_classes = [cls for cls in attrs_classes if attr.has(cls)]
    for klass in attrs_classes:
        # Does the class have this field?
        klass_fields = attr.fields(klass)
        if not hasattr(klass_fields, field_name):
            continue
        field = getattr(klass_fields, field_name)
        metadata = field.metadata
        if PULP2_FIELD in metadata:
            field_name = metadata[PULP2_FIELD]
            converter = metadata.get(PY_PULP2_CONVERTER, lambda x: x)
            return (field_name, matcher._map(converter) if matcher else None)

        # Field was found on the model, but we don't support mapping it to
        # a Pulp field.
        raise NotImplementedError("Searching on field %s is not supported" % field_name)

    # No match => no change, search exactly what was requested
    return None


@attr.s
class PulpSearch(object):
    # Helper class representing a prepared Pulp search.
    # Usually just a filters dict, but may include type_ids
    # for unit searches.
    filters = attr.ib(type=dict)
    type_ids = attr.ib(type=list, default=None)


class TypeIdAccumulator(object):
    def __init__(self):
        self.can_accumulate = True
        self.values = []

    def accumulate_from_match(self, match_expr):
        # Are we still in a state where accumulation is possible?
        if not self.can_accumulate:
            raise ValueError(
                (
                    "Can't serialize criteria for Pulp query; too complicated. "
                    "Try simplifying the query with respect to unit_type/content_type_id."
                )
            )

        # OK, we can accumulate if it's a supported expression type.
        if isinstance(match_expr, dict) and list(match_expr.keys()) == ["$eq"]:
            self.values = [match_expr["$eq"]]
        elif isinstance(match_expr, dict) and list(match_expr.keys()) == ["$in"]:
            self.values = match_expr["$in"]
        else:
            raise ValueError(
                (
                    "Can't serialize criteria for Pulp query, "
                    "unsupported expression for content_type_id: %s\n"
                )
                % repr(match_expr)
            )

        # It is only possible to accumulate once.
        self.can_accumulate = False

    @property
    @contextlib.contextmanager
    def no_accumulate(self):
        old_can_accumulate = self.can_accumulate
        self.can_accumulate = False
        try:
            yield
        finally:
            self.can_accumulate = old_can_accumulate


def search_for_criteria(criteria, type_hint=None, type_ids_accum=None):
    # convert a Criteria object to a PulpSearch with filters as used in the Pulp 2.x API:
    # https://docs.pulpproject.org/dev-guide/conventions/criteria.html#search-criteria
    #
    # type_hint optionally provides the class expected to be found by this search.
    # This can impact the application of certain criteria, e.g. it will affect
    # field mappings looked up by FieldMatchCriteria.
    if criteria is None or isinstance(criteria, TrueCriteria):
        return PulpSearch(filters={})

    type_ids_accum = type_ids_accum or TypeIdAccumulator()

    if isinstance(criteria, AndCriteria):
        clauses = [
            search_for_criteria(c, type_hint, type_ids_accum).filters
            for c in criteria._operands
        ]

        # Empty filters do not affect the result and can be simplified.
        clauses = [c for c in clauses if c != {}]

        # A single clause is the same as no $and at all.
        if len(clauses) == 1:
            filters = clauses[0]
        else:
            filters = {"$and": clauses}

    elif isinstance(criteria, OrCriteria):
        with type_ids_accum.no_accumulate:
            filters = {
                "$or": [
                    search_for_criteria(c, type_hint, type_ids_accum).filters
                    for c in criteria._operands
                ]
            }

    elif isinstance(criteria, FieldMatchCriteria):
        field = criteria._field
        matcher = criteria._matcher

        mapped = map_field_for_type(field, matcher, type_hint)
        if mapped:
            field, matcher = mapped

        match_expr = field_match(matcher)

        # If we are looking at the special _content_type_id field
        # for the purpose of a unit search...
        if field == "_content_type_id" and type_hint is Unit:
            # We should not include this into filters, but instead
            # attempt to accumulate it into type_ids_out.
            # This is because type_ids needs to be serialized into the
            # top-level 'criteria' and not 'filters', and there are
            # additional restrictions on its usage.
            type_ids_accum.accumulate_from_match(match_expr)

            filters = {}
        else:
            filters = {field: match_expr}

    else:
        raise TypeError("Not a criteria: %s" % repr(criteria))

    return PulpSearch(filters=filters, type_ids=type_ids_accum.values[:])


def filters_for_criteria(criteria, type_hint=None):
    return search_for_criteria(criteria, type_hint).filters


def field_match(to_match):
    if isinstance(to_match, RegexMatcher):
        return {"$regex": to_match._pattern}

    if isinstance(to_match, EqMatcher):
        return {"$eq": to_mongo_json(to_match._value)}

    if isinstance(to_match, InMatcher):
        return {"$in": to_mongo_json(to_match._values)}

    if isinstance(to_match, ExistsMatcher):
        return {"$exists": True}

    if isinstance(to_match, LessThanMatcher):
        return {"$lt": to_mongo_json(to_match._value)}

    raise TypeError("Not a matcher: %s" % repr(to_match))



import os
import logging
import traceback

from more_executors.retry import RetryPolicy, ExceptionRetryPolicy

from .errors import TaskFailedException

LOG = logging.getLogger("pubtools.pulplib")

ATTEMPTS = int(os.environ.get("PUBTOOLS_PULPLIB_RETRY_ATTEMPTS", "10"))
SLEEP = float(os.environ.get("PUBTOOLS_PULPLIB_RETRY_SLEEP", "1.0"))
MAX_SLEEP = float(os.environ.get("PUBTOOLS_PULPLIB_RETRY_MAX_SLEEP", "120.0"))


class PulpRetryPolicy(RetryPolicy):
    def __init__(self, max_attempts=ATTEMPTS, sleep=SLEEP, max_sleep=MAX_SLEEP):
        super(PulpRetryPolicy, self).__init__()
        self._max_attempts = max_attempts
        self._delegate = ExceptionRetryPolicy(
            max_attempts=max_attempts, sleep=sleep, max_sleep=max_sleep
        )

    def should_retry(self, attempt, future):
        # This is still using only the default ExceptionRetryPolicy behavior.
        # I expect at some point we will need to fine-tune this and look
        # at the error in more detail in order to decide whether to retry.
        retry = self._delegate.should_retry(attempt, future)

        exception = future.exception()
        if exception and getattr(exception, "response", None) is not None:
            # if returned status code is 404, never retry on that
            if exception.response.status_code == 404:
                return False

        if exception and retry:
            self._log_retry(attempt, future)

        return retry

    def sleep_time(self, attempt, future):
        return self._delegate.sleep_time(attempt, future)

    def _log_retry(self, attempt, future):
        # TODO: it would be nice if we could add some info here about
        # *what* is being retried. But it's a bit of a challenge since
        # future objects themselves don't come with any info on what was
        # done to create the future.
        exception = future.exception()
        LOG.warning(
            "Retrying due to error: %s [%d/%d]%s",
            exception,
            attempt,
            self._max_attempts,
            self._traceback(exception),
            extra={"event": {"type": "pulp-retry"}},
        )

    def _traceback(self, exception):
        out = ""

        if isinstance(exception, TaskFailedException):
            # No point logging traceback of this exception, the message
            # "task failed: <id>" is already all the useful info
            pass
        elif hasattr(exception, "__traceback__"):
            out = "\n" + "".join(
                traceback.format_exception(None, exception, exception.__traceback__)
            )

        return out



import hashlib
import json
import logging
import os
import threading
from functools import partial

import requests
import six
from more_executors import Executors
from more_executors.futures import f_map, f_flat_map, f_return, f_proxy, f_sequence
from six.moves import StringIO

from ..page import Page
from ..criteria import Criteria
from ..model import Repository, MaintenanceReport, Distributor, Unit, Task
from .search import search_for_criteria
from .errors import PulpException
from .poller import TaskPoller
from . import retry


LOG = logging.getLogger("pubtools.pulplib")


class Client(object):
    """A client for the Pulp 2.x API.

    This class provides high-level methods for querying a Pulp server and performing
    actions in Pulp.

    **Usage of futures:**

    Most values are returned as :class:`~concurrent.futures.Future` objects.
    Returned futures are resolved when:

    - a request to Pulp succeeded, if operation is synchronous (e.g. get, search)
    - or a request to Pulp succeeded and any spawned tasks succeeded, if operation
      is asynchronous (e.g. publish)

    If a future is currently awaiting one or more Pulp tasks, cancelling the future
    will attempt to cancel those tasks.

    **Client lifecycle:**

    .. versionadded:: 2.12.0

    Client instances support the context manager protocol and can be used
    via a ``with`` statement, as in example:

    .. code-block:: python

        with Client(url="https://pulp.example.com/") as client:
            do_something_with(client)

    While not mandatory, it is encouraged to ensure that any threads associated with
    the client are promptly shut down.

    **Proxy futures:**

    .. versionadded:: 2.1.0

    All :class:`~concurrent.futures.Future` objects produced by this client are
    *proxy futures*, meaning that attribute and method lookups on the objects are
    proxied to the future's result, blocking as needed.

    This allows the client to be used within blocking code without having to
    scatter calls to ``.result()`` throughout.

    For example, this block of code:

    .. code-block:: python

        repo = client.get_repository(repo_id).result()
        publish_tasks = repo.publish().result()
        task_ids = ','.join([t.id for t in publish_tasks])
        log.info("Published %s: %s", repo.id, task_ids)

    ...may be alternatively written without the calls to ``.result()``, due to
    the usage of proxy futures:

    .. code-block:: python

        repo = client.get_repository(repo_id)
        publish_tasks = repo.publish()
        task_ids = ','.join([t.id for t in publish_tasks])
        log.info("Published %s: %s", repo.id, task_ids)

    **Retries:**

    In general, for all methods which represent an idempotent operation and
    which return a Future, the library may retry operations several times in case
    of failure.

    **Throttling:**

    This client will internally throttle the number of Pulp tasks allowed to be
    outstanding at any given time, so that no single client can overload the Pulp
    task queue.
    """

    # Various defaults can be controlled by environment variables.
    # These are not a documented/supported feature of the library.
    # They're more of an emergency escape hatch to temporarily resolve issues.
    _REQUEST_THREADS = int(os.environ.get("PUBTOOLS_PULPLIB_REQUEST_THREADS", "4"))
    _PAGE_SIZE = int(os.environ.get("PUBTOOLS_PULPLIB_PAGE_SIZE", "2000"))
    _TASK_THROTTLE = int(os.environ.get("PUBTOOLS_PULPLIB_TASK_THROTTLE", "200"))
    _CHUNK_SIZE = int(os.environ.get("PUBTOOLS_PULPLIB_CHUNK_SIZE", 1024 * 1024))

    # Policy used when deciding whether to retry operations.
    # This is mainly provided here as a hook for autotests, so the policy can be
    # overridden there.
    _RETRY_POLICY = retry.PulpRetryPolicy()

    def __init__(self, url, **kwargs):
        """
        Args:
            str url
                URL of a Pulp server, e.g. https://pulp.example.com/.

            int task_throttle
                Maximum number of queued or running tasks permitted for this client.
                If more than this number of tasks are running, the client will wait before triggering more.
                This can be used to ensure no single client overwhelms the Pulp server.

            object auth, cert, headers, max_redirects, params, proxies, verify
                Any of these arguments, if provided, are used to initialize
                :class:`requests.Session` objects used by the client.

                These may be used, for example, to configure the credentials
                for the Pulp server or to use an alternative CA bundle.
        """
        self._url = url

        while self._url.endswith("/"):
            self._url = self._url[:-1]

        self._session_kwargs = {}
        for arg in (
            "auth",
            "cert",
            "headers",
            "max_redirects",
            "params",
            "proxies",
            "verify",
        ):
            if arg in kwargs:
                self._session_kwargs[arg] = kwargs.pop(arg)

        _task_throttle = kwargs.pop("task_throttle", self._TASK_THROTTLE)

        if kwargs:
            raise TypeError(
                "Unexpected keyword argument(s) %s" % ",".join(kwargs.keys())
            )

        self._tls = threading.local()

        # Cache supported types, used during searches.
        self._server_type_ids = None

        # executor only for issuing HTTP requests to Pulp:
        # - does not cover watching Pulp tasks
        # - checks HTTP response and raises if it's not JSON or
        #   it's not successful
        self._request_executor = (
            Executors.thread_pool(
                name="pubtools-pulplib-requests", max_workers=self._REQUEST_THREADS
            )
            .with_map(self._unpack_response)
            .with_retry(retry_policy=self._RETRY_POLICY)
        )

        # executor for issuing HTTP requests to Pulp which spawn tasks:
        # - checks HTTP response and raises if it's not JSON or
        #   it's not successful
        # - unpacks spawned_tasks from response to find task IDs to watch
        # - waits for tasks to complete
        # - retries whole thing (resubmitting request & making a new task) if needed
        # - throttles number of tasks pending
        poller = TaskPoller(self._new_session(), self._url)
        self._task_executor = (
            Executors.thread_pool(
                name="pubtools-pulplib-tasks", max_workers=self._REQUEST_THREADS
            )
            .with_map(self._unpack_response)
            .with_map(self._log_spawned_tasks)
            .with_poll(poller, cancel_fn=poller.cancel)
            .with_throttle(_task_throttle)
            .with_retry(retry_policy=self._RETRY_POLICY)
        )

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self._request_executor.__exit__(*args, **kwargs)
        self._task_executor.__exit__(*args, **kwargs)

    def get_repository(self, repository_id):
        """Get a repository by ID.

        Args:
            repository_id (str)
                The ID of the repository to be fetched.

        Returns:
            Future[:class:`~pubtools.pulplib.Repository`]
                A future holding the fetched Repository.
        """
        page_f = self.search_repository(Criteria.with_id(repository_id))

        def unpack_page(page):
            if len(page.data) != 1:
                raise PulpException("Repository id=%s was not found" % repository_id)
            return page.data[0]

        repo_f = f_map(page_f, unpack_page)
        return f_proxy(repo_f)

    def search_repository(self, criteria=None):
        """Search for repositories matching the given criteria.

        Args:
            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used for this search.
                If None, search for all repositories.

        Returns:
            Future[:class:`~pubtools.pulplib.Page`]
                A future representing the first page of results.

                Each page will contain a collection of
                :class:`~pubtools.pulplib.Repository` objects.
        """
        search_options = {"distributors": True}
        return self._search(
            Repository, "repositories", criteria=criteria, search_options=search_options
        )

    def search_content(self, criteria=None):
        """Search for units across all repositories.

        Args:
            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used for this search.
                If None, search for all units.

        Returns:
            Future[:class:`~pubtools.pulplib.Page`]
                A future representing the first page of results.

                Each page will contain a collection of
                :class:`~pubtools.pulplib.Unit` subclasses objects.

        .. versionadded:: 2.6.0
        """
        # Criteria will be serialized into a Pulp search at the time we
        # actually do the query, but validate eagerly as well so we raise
        # ASAP on invalid input.
        search_for_criteria(criteria, Unit, None)

        if self._server_type_ids is None:
            # We'll be using this in a moment either to set default IDs or to
            # reject searches for invalid types.
            # Note: no locking, so if we're called from multiple threads we in
            # theory might waste some time querying the types more than once.
            self._server_type_ids = self.get_content_type_ids()

        return f_proxy(
            f_flat_map(
                self._server_type_ids,
                lambda ids: self._search_content_with_server_type_ids(criteria, ids),
            )
        )

    def copy_content(self, from_repository, to_repository, criteria=None):
        """Copy content from one repository to another.

        Args:
            from_repository (:class:`~pubtools.pulplib.Repository`)
                A repository used as the source of content.

            to_repository (:class:`~pubtools.pulplib.Repository`)
                A repository to receive copied content.

            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used to find units for copy.
                If None, all units in the source repo will be copied.

        Returns:
            Future[list[:class:`~pubtools.pulplib.Task`]]
                A future which is resolved when the copy completes.

                The :meth:`~pubtools.pulplib.Task.units` attribute may be inspected
                to determine which unit(s) were copied. Note that the returned units
                typically will have only a subset of available fields.

        .. versionadded:: 2.17.0
        """
        return f_proxy(
            self._do_associate(from_repository.id, to_repository.id, criteria)
        )

    def update_content(self, unit):
        """Update mutable fields on an existing unit.

        Args:
            unit (:class:`~pubtools.pulplib.Unit`)
                A unit to be updated.

                This unit must have a known ``unit_id``.

                Only those fields documented as *mutable* will have any effect during
                the update (e.g. ``cdn_path``). Other fields either cannot be updated
                at all, or can only be updated by re-uploading the associated content.

        Returns:
            Future
                A future which is resolved with a value of ``None`` once the unit
                has been updated.

        .. versionadded:: 2.20.0
        """

        if not unit.unit_id:
            raise ValueError("unit_id missing on call to update_content()")

        url = os.path.join(
            self._url,
            "pulp/api/v2/content/units/%s/%s/pulp_user_metadata/"
            % (unit.content_type_id, unit.unit_id),
        )

        out = self._request_executor.submit(
            self._do_request, method="PUT", url=url, json=unit._usermeta
        )

        # The Pulp API is defined as returning 'null' so this should be a no-op,
        # but to be extra sure we don't return anything unexpected, we'll force
        # the return value to None.
        out = f_map(out, lambda _: None)

        return out

    def _search_content_with_server_type_ids(self, criteria, server_type_ids):
        prepared_search = search_for_criteria(criteria, Unit, None)
        type_ids = prepared_search.type_ids
        if not type_ids:
            type_ids = server_type_ids
        missing = set(type_ids) - set(server_type_ids)
        if missing:
            raise ValueError(
                "Content type: %s is not supported by server" % ",".join(missing)
            )

        return self._search(
            Unit,
            ["content/units/%s" % x for x in type_ids],
            criteria=criteria,
            search_options={"include_repos": True},
        )

    def search_distributor(self, criteria=None):
        """Search the distributors matching the given criteria.

        Args:
            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used for this search.
                If None, search for all distributors.

        Returns:
            Future[:class:`~pubtools.pulplib.Page`]
                A future representing the first page of results.

                Each page will contain a collection of
                :class:`~pubtools.pulplib.Distributor` objects.

        .. versionadded:: 2.1.0
        """
        return self._search(Distributor, "distributors", criteria=criteria)

    def search_task(self, criteria=None):
        """Search the tasks matching the given criteria.

        Args:
            criteria (:class:`~pubtools.pulplib.Criteria`)
                A criteria object used for this search.
                If None, search for all tasks.

        Returns:
            Future[:class:`~pubtools.pulplib.Page`]
                A future representing the first page of results.

                Each page will contain a collection of
                :class:`~pubtools.pulplib.Task` objects.

        .. versionadded:: 2.19.0
        """
        return self._search(Task, "tasks", criteria=criteria)

    def _search(
        self,
        return_type,
        resource_types,
        search_type="search",
        search_options=None,
        criteria=None,
    ):  # pylint:disable = too-many-arguments

        if not isinstance(resource_types, (list, tuple)):
            resource_types = [resource_types]

        responses = []
        searches = []
        urls = []
        for resource_type in resource_types:
            url = os.path.join(
                self._url, "pulp/api/v2/%s/%s/" % (resource_type, search_type)
            )
            urls.append(url)
            prepared_search = search_for_criteria(criteria, return_type)

            search = {
                "criteria": {
                    "skip": 0,
                    "limit": self._PAGE_SIZE,
                    "filters": prepared_search.filters,
                }
            }
            search.update(search_options or {})

            if search_type == "search/units":
                # Unit searches need a little special handling:
                # - serialization might have extracted some type_ids
                # - filters should be wrapped under 'unit'
                #   (we do not support searching on associations right now)
                if prepared_search.type_ids:
                    search["criteria"]["type_ids"] = prepared_search.type_ids
                search["criteria"]["filters"] = {"unit": search["criteria"]["filters"]}

            searches.append(search)
            responses.append(self._do_search(url, search))

        # When this request is resolved, we'll have the first page of data.
        # We'll need to convert that into a page and also keep going with
        # the search if there's more to be done.
        return f_proxy(
            f_map(
                f_sequence(responses),
                lambda data: self._handle_page(urls, return_type, searches, data),
            )
        )

    def _search_repo_units(self, repo_id, criteria):
        resource_type = "repositories/%s" % repo_id

        return self._search(
            Unit, resource_type, search_type="search/units", criteria=criteria
        )

    def get_maintenance_report(self):
        """Get the current maintenance mode status for this Pulp server.

        Returns:
            Future[:class:`~pubtools.pulplib.MaintenanceReport`]
                A future describes the maintenance status

        .. versionadded:: 1.4.0
        """
        report_ft = self._do_get_maintenance()

        return f_map(
            report_ft,
            lambda data: MaintenanceReport._from_data(data)
            if data
            else MaintenanceReport(),
        )

    def set_maintenance(self, report):
        """Set maintenance mode for this Pulp server.

        Args:
            report:
                An updated :class:`~pubtools.pulplib.MaintenanceReport` object that
                will be used as the newest maintenance report.
        Return:
            Future[list[:class:`~pubtools.pulplib.Task`]]
                A future which is resolved when maintenance mode has been updated successfully.

                The future contains a task triggered and awaited during the publish
                maintenance repository operation.

        .. versionadded:: 1.4.0
        """
        report_json = json.dumps(report._export_dict(), indent=4, sort_keys=True)
        report_fileobj = StringIO(report_json)

        repo = self.get_repository("redhat-maintenance").result()

        # upload updated report to repository and publish
        upload_ft = repo.upload_file(report_fileobj, "repos.json")

        return f_flat_map(upload_ft, lambda _: repo.publish())

    def get_content_type_ids(self):
        """Get the content types supported by this Pulp server.

        Returns:
            Future[list[str]]
                A future holding the IDs of supported content types.

                The returned values will depend on the plugins installed
                on the connected Pulp server.

                "modulemd", "rpm", "srpm" and "erratum" are some examples
                of common return values.

        .. versionadded:: 1.4.0
        """
        url = os.path.join(self._url, "pulp/api/v2/plugins/types/")

        out = self._request_executor.submit(self._do_request, method="GET", url=url)

        # The pulp API returns an object per supported type.
        # We only support returning the ID at this time.
        return f_proxy(f_map(out, lambda types: sorted([t["id"] for t in types])))

    def _do_upload_file(self, upload_id, file_obj):
        def do_next_upload(checksum, size):
            data = file_obj.read(self._CHUNK_SIZE)
            if data:
                if isinstance(data, six.text_type):
                    # if it's unicode, need to encode before calculate checksum
                    data = data.encode("utf-8")
                checksum.update(data)
                return f_flat_map(
                    self._do_upload(data, upload_id, size),
                    lambda _: do_next_upload(checksum, size + len(data)),
                )
            # nothing more to upload, return checksum and size
            return f_return((checksum.hexdigest(), size))

        is_file_object = "close" in dir(file_obj)
        if not is_file_object:
            file_obj = open(file_obj, "rb")

        upload_f = f_flat_map(f_return(), lambda _: do_next_upload(hashlib.sha256(), 0))

        upload_f.add_done_callback(lambda _: file_obj.close())

        return upload_f

    def _publish_repository(self, repo, distributors_with_config):
        tasks_f = f_return([])

        def do_next_publish(accumulated_tasks, distributor, config):
            distributor_tasks_f = self._publish_distributor(
                repo.id, distributor.id, config
            )
            return f_map(
                distributor_tasks_f,
                lambda distributor_tasks: accumulated_tasks + distributor_tasks,
            )

        for (distributor, config) in distributors_with_config:
            next_publish = partial(
                do_next_publish, distributor=distributor, config=config
            )
            tasks_f = f_flat_map(tasks_f, next_publish)

        return tasks_f

    def _publish_distributor(self, repo_id, distributor_id, override_config):
        url = os.path.join(
            self._url, "pulp/api/v2/repositories/%s/actions/publish/" % repo_id
        )
        body = {"id": distributor_id, "override_config": override_config}
        return self._task_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )

    def _do_associate(self, src_repo_id, dest_repo_id, criteria=None):
        url = os.path.join(
            self._url, "pulp/api/v2/repositories/%s/actions/associate/" % dest_repo_id
        )

        pulp_search = search_for_criteria(criteria, type_hint=Unit, type_ids_accum=None)

        body = {"source_repo_id": src_repo_id, "criteria": {}}
        if pulp_search.type_ids:
            body["criteria"]["type_ids"] = pulp_search.type_ids
        if pulp_search.filters:
            body["criteria"]["filters"] = {"unit": pulp_search.filters}

        LOG.debug("Submitting %s associate: %s", url, body)

        return self._task_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )

    def _do_unassociate(self, repo_id, criteria=None):
        url = os.path.join(
            self._url, "pulp/api/v2/repositories/%s/actions/unassociate/" % repo_id
        )

        # use type hint=Unit so that if type_ids are the goal here
        # then we will get back a properly prepared PulpSearch with
        # a populated type_ids field
        pulp_search = search_for_criteria(criteria, type_hint=Unit, type_ids_accum=None)

        body = {"criteria": {"type_ids": pulp_search.type_ids}}

        LOG.debug("Submitting %s unassociate: %s", url, body)

        return self._task_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )

    @classmethod
    def _unpack_response(cls, pulp_response):
        try:
            parsed = pulp_response.json()
        except Exception:
            # Couldn't parse as JSON?
            # If the response was unsuccessful, raise that.
            # Otherwise re-raise parse error.
            pulp_response.raise_for_status()
            raise

        if (
            isinstance(parsed, dict)
            and parsed.get("http_status") == 404
            and parsed.get("http_request_method") == "DELETE"
        ):
            # Special case allowing unsuccessful status:
            # If we asked to DELETE something, and we got a 404 response,
            # this is not considered an error because the postcondition
            # of our request is satisfied.
            # Hence, don't call raise_for_status.
            pass
        else:
            # In general, we'll raise if response was unsuccessful.
            pulp_response.raise_for_status()

        return parsed

    @classmethod
    def _log_spawned_tasks(cls, taskdata):
        try:
            spawned = taskdata.get("spawned_tasks") or []
            for task in spawned:
                LOG.info("Created Pulp task: %s", task["task_id"])
        except Exception:  # pylint: disable=broad-except
            # something wrong with the data, can't log.
            # This error will be raised elsewhere
            pass
        return taskdata

    def _handle_page(self, urls, object_class, searches, raw_data_list):

        page_data = []
        for url, search, raw_data in zip(urls, searches, raw_data_list):
            # Extract metadata from Pulp units
            LOG.debug("Got pulp response for %s, %s elems", search, len(raw_data))
            if object_class is Unit and url.endswith("units/"):
                raw_data = [elem["metadata"] for elem in raw_data]

            page_data.extend([object_class.from_data(elem) for elem in raw_data])

        for obj in page_data:
            # set_client is only applicable for repository and distributor objects
            if hasattr(obj, "_set_client"):
                obj._set_client(self)

        # Do we need a next page?
        next_page = None

        next_responses = []
        next_searches = []
        next_urls = []
        for url, search, raw_data in zip(urls, searches, raw_data_list):
            limit = search["criteria"]["limit"]
            if len(raw_data) == limit:
                # Yeah, we might...
                search = search.copy()
                search["criteria"] = search["criteria"].copy()
                search["criteria"]["skip"] = search["criteria"]["skip"] + limit
                response_f = self._do_search(url, search)
                next_searches.append(search)
                next_responses.append(response_f)
                next_urls.append(url)

        if next_responses:
            next_page = f_proxy(
                f_map(
                    f_sequence(next_responses),
                    lambda data: self._handle_page(
                        next_urls, object_class, next_searches, data
                    ),
                )
            )
        return Page(data=page_data, next=next_page)

    @property
    def _session(self):
        if not hasattr(self._tls, "session"):
            self._tls.session = self._new_session()
        return self._tls.session

    def _new_session(self):
        out = requests.Session()
        for key, value in self._session_kwargs.items():
            setattr(out, key, value)
        return out

    def _do_request(self, **kwargs):
        return self._session.request(**kwargs)

    def _do_search(self, url, search):
        LOG.debug("Submitting %s search: %s", url, search)
        return self._request_executor.submit(
            self._do_request, method="POST", url=url, json=search
        )

    def _delete_resource(self, resource_type, resource_id):
        url = os.path.join(
            self._url, "pulp/api/v2/%s/%s/" % (resource_type, resource_id)
        )

        LOG.debug("Queuing request to DELETE %s", url)
        return self._task_executor.submit(self._do_request, method="DELETE", url=url)

    def _request_upload(self):
        url = os.path.join(self._url, "pulp/api/v2/content/uploads/")

        LOG.debug("Requesting upload id")
        return self._request_executor.submit(self._do_request, method="POST", url=url)

    def _do_upload(self, data, upload_id, offset):
        url = os.path.join(
            self._url, "pulp/api/v2/content/uploads/%s/%s/" % (upload_id, offset)
        )

        return self._request_executor.submit(
            self._do_request, method="PUT", url=url, data=data
        )

    def _do_import(
        self, repo_id, upload_id, unit_type_id, unit_key, unit_metadata=None
    ):
        url = os.path.join(
            self._url, "pulp/api/v2/repositories/%s/actions/import_upload/" % repo_id
        )

        unit_metadata = unit_metadata or {}

        body = {
            "unit_type_id": unit_type_id,
            "upload_id": upload_id,
            "unit_key": unit_key,
            "unit_metadata": unit_metadata,
        }

        LOG.debug("Importing contents to repo %s with upload id %s", repo_id, upload_id)
        return self._task_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )

    def _delete_upload_request(self, upload_id):
        url = os.path.join(self._url, "pulp/api/v2/content/uploads/%s/" % upload_id)

        LOG.debug("Deleting upload request %s", upload_id)
        return self._request_executor.submit(self._do_request, method="DELETE", url=url)

    def _do_get_maintenance(self):
        def map_404_to_none(exception):
            # Translates 404 errors to a None response (no maintenance report).
            if (
                getattr(exception, "response", None) is not None
                and exception.response.status_code == 404
            ):
                return None
            # Any other types of errors are raised unchanged.
            raise exception

        url = os.path.join(self._url, "pulp/isos/redhat-maintenance/repos.json")

        response = self._request_executor.submit(
            self._do_request, method="GET", url=url
        )

        return f_map(response, error_fn=map_404_to_none)

    def _do_sync(self, repo_id, sync_options):
        if not sync_options["feed"]:
            raise ValueError("Cannot sync with empty feed: '%s'" % sync_options["feed"])

        url = os.path.join(
            self._url, "pulp/api/v2/repositories/%s/actions/sync/" % repo_id
        )

        body = {"override_config": sync_options}

        LOG.debug("Syncing repository %s with feed %s", repo_id, sync_options["feed"])
        return self._task_executor.submit(
            self._do_request, method="POST", url=url, json=body
        )



import logging
import datetime
import os

import jsonschema
from frozenlist2 import frozenlist


from pubtools.pulplib._impl import compat_attr as attr
from .attr import pulp_attrib
from .convert import read_timestamp, write_timestamp
from ..schema import load_schema
from .common import InvalidDataException


LOG = logging.getLogger("pubtools.pulplib")


USER = os.environ.get("USER")
HOSTNAME = os.environ.get("HOSTNAME")


@attr.s(kw_only=True, frozen=True)
class MaintenanceEntry(object):
    """Details about the maintenance status of a specific repository.

    .. versionadded:: 1.4.0
    """

    repo_id = pulp_attrib(type=str)
    """ID of repository in maintenance.

    Note: there is no guarantee that a repository of this ID currently exists
    in the Pulp server."""
    message = pulp_attrib(default=None, type=str)
    """Why this repository is in maintenance."""
    owner = pulp_attrib(default=None, type=str)
    """Who set this repository in maintenance mode."""
    started = pulp_attrib(default=None, type=datetime.datetime)
    """:class:`~datetime.datetime` in UTC at when the maintenance started."""


@attr.s(kw_only=True, frozen=True)
class MaintenanceReport(object):
    """Represents the maintenance status of Pulp repositories.

    On release-engineering Pulp servers, it's possible to put individual repositories
    into "maintenance mode".  When in maintenance mode, external publishes of a repository
    will be blocked.  Other operations remain possible.

    This object holds information on the set of repositories currently in maintenance mode.

    .. versionadded:: 1.4.0
    """

    _OWNER = "%s@%s" % (USER, HOSTNAME) if all([USER, HOSTNAME]) else "pubtools.pulplib"

    _SCHEMA = load_schema("maintenance")

    last_updated = pulp_attrib(default=None, type=datetime.datetime)
    """:class:`~datetime.datetime` in UTC when this report was last updated,
    if it's the first time the report is created, current time is used."""

    last_updated_by = pulp_attrib(default=None, type=str)
    """Person/party who updated the report last time."""

    entries = pulp_attrib(
        default=attr.Factory(frozenlist), type=list, converter=frozenlist
    )
    """A list of :class:`MaintenanceEntry` objects, indicating
    which repositories are in maintenance mode and details.
    If empty, then it means no repositories are in maintenance mode.
    """

    @entries.validator
    def _check_duplicates(self, _, value):
        # check if there's duplicate entries
        repo_ids = [entry.repo_id for entry in value]
        if len(repo_ids) != len(set(repo_ids)):
            raise ValueError("Duplicate entries")

    @classmethod
    def _from_data(cls, data):
        """Create a new report with raw data

        Args:
            data (dict):
                A dict containing a raw representation of the maintenance status.

        Returns:
            a new instance of ``cls``

        Raises:
            InvalidDataException
                If the provided ``data`` fails validation against an expected schema.
        """
        try:
            jsonschema.validate(instance=data, schema=cls._SCHEMA)
        except jsonschema.exceptions.ValidationError as error:
            LOG.exception("%s.from_data invoked with invalid data", cls.__name__)
            raise InvalidDataException(str(error))

        entries = []
        for repo_id, details in data["repos"].items():
            entries.append(
                MaintenanceEntry(
                    repo_id=repo_id,
                    message=details["message"],
                    owner=details["owner"],
                    started=read_timestamp(details["started"]),
                )
            )

        maintenance = cls(
            last_updated=read_timestamp(data["last_updated"]),
            last_updated_by=data["last_updated_by"],
            entries=entries,
        )

        return maintenance

    def _export_dict(self):
        """export a raw dictionary of maintenance report"""
        report = {
            "last_updated": write_timestamp(self.last_updated),
            "last_updated_by": self.last_updated_by or self._OWNER,
            "repos": {},
        }

        for entry in self.entries:
            report["repos"].update(
                {
                    entry.repo_id: {
                        "message": entry.message,
                        "owner": entry.owner,
                        "started": write_timestamp(entry.started),
                    }
                }
            )

        return report

    def add(self, repo_ids, **kwargs):
        """Add entries to maintenance report and update the timestamp. Every
        entry added to the report represents a repository in maintenance mode.

        Args:
            repo_ids (list[str]):
                A list of repository ids. New entries with these repository ids will
                be added to the maintenance report.

                Note: it's users' responsibility to make sure the repository exists in
                the Pulp server, this method doesn't check for the existence of repositories.

            message (str) (optional):
                Reason why put the repo to maintenance.

            owner (str) (optional):
                Who set the maintenance mode.

        Returns:
           :class:`~pubtools.pulplib.MaintenanceReport`
                A copy of this maintenance report with added repositories.

        """
        message = kwargs.get("message") or "Maintenance mode is enabled"
        owner = kwargs.get("owner") or self._OWNER

        to_add = []
        for repo in repo_ids:
            to_add.append(
                MaintenanceEntry(
                    repo_id=repo,
                    owner=owner,
                    message=message,
                    started=datetime.datetime.utcnow(),
                )
            )
        entries = list(self.entries)
        entries.extend(to_add)

        # filter out duplicated entries. Filtering is in reverse order, which
        # means existed entries will be replaced by newer ones with same repo_id
        filtered_entries = []
        entry_ids = set()
        for entry in reversed(entries):
            if entry.repo_id not in entry_ids:
                filtered_entries.append(entry)
                entry_ids.add(entry.repo_id)

        return attr.evolve(
            self,
            entries=filtered_entries,
            last_updated_by=owner,
            last_updated=datetime.datetime.utcnow(),
        )

    def remove(self, repo_ids, **kwargs):
        """Remove entries from the maintenance report. Remove entries means the
        removing corresponding repositories from maintenance mode.

        Args:
            repo_ids (list[str]):
                A list of repository ids. Entries match repository ids will be removed
                from the maintenance report.

            owner (str) (optional):
                Who unset the maintenance mode.

        Returns:
            :class:`~pubtools.pulplib.MaintenanceReport`
                A copy of this maintenance report with removed repositories.
        """
        owner = kwargs.get("owner") or self._OWNER

        repo_ids = set(repo_ids)
        # convert to set, make checking faster
        new_entries = []
        for entry in self.entries:
            if entry.repo_id not in repo_ids:
                new_entries.append(entry)

        return attr.evolve(self, last_updated_by=owner, entries=new_entries)



# This file contains tests for the model-related assertions themselves
# to prove that the assertions aren't simply passing without verifying anything.
import attr
from attr import validators
import pytest

from .assertions import assert_model_invariants


def test_wrong_type():
    """An object with an attribute not matching defined 'type' should fail tests."""

    @attr.s(frozen=True)
    class BadType(object):
        id = attr.ib(type=str)

    instance = BadType(id=12345)

    with pytest.raises(AssertionError) as exc:
        assert_model_invariants(instance)

    assert "BadType.id should be a str but is not" in str(exc.value)


def test_unvalidated_type():
    """An object with an attribute with unvalidated type should fail"""

    @attr.s(frozen=True)
    class Unvalidated(object):
        my_attr = attr.ib(type=str)

    instance = Unvalidated(my_attr="foo")

    with pytest.raises(AssertionError) as exc:
        assert_model_invariants(instance)

    assert "copying with wrong type for my_attr was possible" in str(exc.value)


def test_bad_hash():
    """An object which is not hashable should fail tests."""

    @attr.s(frozen=True, hash=False)
    class BadHash(object):
        id = attr.ib(type=str, validator=validators.instance_of(str))

        def __hash__(self):
            raise TypeError("Can't hash")

    instance = BadHash(id="bad-hash")

    with pytest.raises(TypeError):
        assert_model_invariants(instance)


def test_bad_copied_hash():
    """An object whose copy has a different hash value will fail tests."""

    callnum = []

    @attr.s(frozen=True, hash=False)
    class BadHashOnCopy(object):
        id = attr.ib(type=str, validator=validators.instance_of(str), hash=True)

        def __hash__(self):
            callnum.append(None)
            return hash(tuple(callnum))

    instance = BadHashOnCopy(id="some-object")

    with pytest.raises(AssertionError) as exc:
        assert_model_invariants(instance)

    assert "hash differs" in str(exc.value)


def test_bad_copied_equality():
    """An object whose copy is not equal to the original will fail tests."""

    @attr.s(init=False, hash=False, frozen=True)
    class BadCopy(object):
        some_attr = attr.ib(
            default=None,
            type=str,
            validator=validators.optional(validators.instance_of(str)),
        )

        def __init__(self, some_attr):
            # Handle this argument oddly so that copying this class
            # doesn't actually get you a plain copy
            self.__dict__["some_attr"] = some_attr + "-suffix"

        def __hash__(self):
            # to make it get past the hash check
            return 0

    instance = BadCopy(some_attr="abc")

    with pytest.raises(AssertionError) as exc:
        assert_model_invariants(instance)

    assert "copied object is not equal" in str(exc.value)



from setuptools import setup, find_packages


def get_description():
    return "A Pulp library for publishing tools"


def get_long_description():
    with open("README.md") as f:
        text = f.read()

    # Long description is everything after README's initial heading
    idx = text.find("\n\n")
    return text[idx:]


def get_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


setup(
    name="pubtools-pulplib",
    version="2.22.0",
    packages=find_packages(exclude=["tests"]),
    package_data={"pubtools.pulplib._impl.schema": ["*.yaml"]},
    url="https://github.com/release-engineering/pubtools-pulplib",
    license="GNU General Public License",
    description=get_description(),
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=get_requirements(),
    python_requires=">=2.6",
    project_urls={
        "Documentation": "https://release-engineering.github.io/pubtools-pulplib/",
        "Changelog": "https://github.com/release-engineering/pubtools-pulplib/blob/master/CHANGELOG.md",
    },
)



import logging
import weakref
import warnings
from concurrent.futures import Future

from frozenlist2 import frozenlist


from . import compat_attr as attr
from .model.attr import pulp_attrib

LOG = logging.getLogger("pubtools.pulplib")


# pylint shows these spurious errors on this file, not sure why:
# page.py:86:23: E1101: Instance of '_CountingAttr' has no 'done' member (no-member)
# Could be due to bug https://github.com/PyCQA/pylint/issues/1694 ?
# pylint: disable=no-member


@attr.s(kw_only=True, frozen=True)
class Page(object):
    """A page of Pulp search results.

    All Pulp searches issued by this library are paginated.
    Instances of this class may be used to iterate through the returned
    pages, in both blocking and non-blocking coding styles.

    Page objects are iterables. Iterating over a page will iterate
    over all data within that page, *and all subsequent pages*,
    blocking on more data from Pulp as needed.

    Examples:

        **Non-blocking iteration**

        This example sets up non-blocking processing of all results from
        a search query. Futures are chained and composed using
        :func:`~more_executors.futures.f_flat_map` and friends.

        .. code-block:: python

            def handle_results(page):
                # Returns a future for handling of a single page
                for repo in page.data:
                    do_something(repo)
                if page.next:
                    # There's more data, handle that too when it's ready
                    return f_flat_map(page.next, handle_results)
                # No more data, we're done
                print("Handled results!")
                return f_return()

            page_f = client.search_repository(...)
            handled_f = f_flat_map(page_f, handle_results)
            # handled_f will be resolved when all results are handled

        **Blocking iteration**

        This example uses the page as an iterable to loop over all search
        results. At certain points during the iteration, blocking may
        occur to await more pages from Pulp.

        .. code-block:: python

            page = client.search_repository(...).result()
            # processes all data, but may block at page boundaries
            for repo in page:
                do_something(repo)

    """

    data = pulp_attrib(
        default=attr.Factory(frozenlist), type=list, converter=frozenlist
    )
    """List of Pulp objects in this page.

    This list will contain instances of the appropriate Pulp object type
    (e.g. :class:`~pubtools.pulplib.Repository` for a repository search).
    """

    next = pulp_attrib(default=None, type=Future)
    """None, if this is the last page of results.

    Otherwise, a Future[:class:`Page`] for the next page of results.
    """

    def __attrs_post_init__(self):
        if self.next:
            # If *this* page disappears, it's pointless to keep querying for
            # the next page, so we could cancel that future.
            #
            # Note: when we are Py3-only, this could be better rewritten
            # using weakref.finalize or even just plain __del__

            to_cancel = self.next

            def do_cancel(*_):
                if not to_cancel.done():
                    cancel_result = to_cancel.cancel()
                    LOG.debug("Cancel next page due to GC: %s", cancel_result)

            # Just stash a weakref anywhere it'll stay alive for as long
            # as the page itself.
            self.__dict__["_cancel_ref"] = weakref.ref(self, do_cancel)

    def as_iter(self):
        # TODO: remove me. Originally deprecated 2019-09.
        warnings.warn(
            "as_iter is deprecated, use page as iterable instead", DeprecationWarning
        )

        return self.__iter__()

    def __iter__(self):
        page = self
        while True:
            for elem in page.data:
                yield elem
            if not page.next:
                return
            # blocks here if next page is not yet fetched
            page = page.next.result()



import os
import logging

import yaml
import jsonschema


LOG = logging.getLogger("pubtools.pulplib")


def load_schema(basename, definition=None):
    """Load a JSON schema file from YAML.

    If definition is given, a named subschema can be referenced from
    'definitions' section.
    """

    filename = os.path.join(os.path.dirname(__file__), "%s.yaml" % basename)
    with open(filename) as file:
        schema = yaml.safe_load(file)

    if definition:
        schema = schema["definitions"][definition]

    return schema
